from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
import json
import os
from pathlib import Path
from typing import Callable

from skillos.storage import atomic_write_text, file_lock
from skillos.tenancy import resolve_tenant_root


DEFAULT_FAILURE_THRESHOLD = 5
DEFAULT_WINDOW_SECONDS = 300
DEFAULT_OPEN_SECONDS = 300
DEFAULT_HALF_OPEN_MAX = 1


def default_circuit_breaker_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "runtime" / "circuit_breaker.json"


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _ensure_utc(value: datetime) -> datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=timezone.utc)
    return value.astimezone(timezone.utc)


def _format_datetime(value: datetime | None) -> str | None:
    if value is None:
        return None
    return _ensure_utc(value).isoformat()


def _parse_datetime(value: str | None) -> datetime | None:
    if not value:
        return None
    parsed = datetime.fromisoformat(value)
    return _ensure_utc(parsed)


@dataclass(frozen=True)
class CircuitBreakerConfig:
    failure_threshold: int = DEFAULT_FAILURE_THRESHOLD
    window_seconds: int = DEFAULT_WINDOW_SECONDS
    open_seconds: int = DEFAULT_OPEN_SECONDS
    half_open_max_attempts: int = DEFAULT_HALF_OPEN_MAX


def circuit_breaker_config_from_env() -> CircuitBreakerConfig:
    return CircuitBreakerConfig(
        failure_threshold=_env_int(
            "SKILLOS_CIRCUIT_FAILURE_THRESHOLD",
            DEFAULT_FAILURE_THRESHOLD,
        ),
        window_seconds=_env_int(
            "SKILLOS_CIRCUIT_WINDOW_SECONDS",
            DEFAULT_WINDOW_SECONDS,
        ),
        open_seconds=_env_int(
            "SKILLOS_CIRCUIT_OPEN_SECONDS",
            DEFAULT_OPEN_SECONDS,
        ),
        half_open_max_attempts=_env_int(
            "SKILLOS_CIRCUIT_HALF_OPEN_MAX",
            DEFAULT_HALF_OPEN_MAX,
        ),
    )


@dataclass
class CircuitState:
    state: str = "closed"
    failure_count: int = 0
    last_failure_at: datetime | None = None
    opened_at: datetime | None = None
    half_open_attempts: int = 0

    def to_dict(self) -> dict[str, object]:
        return {
            "state": self.state,
            "failure_count": self.failure_count,
            "last_failure_at": _format_datetime(self.last_failure_at),
            "opened_at": _format_datetime(self.opened_at),
            "half_open_attempts": self.half_open_attempts,
        }

    @classmethod
    def from_dict(cls, payload: dict[str, object]) -> "CircuitState":
        return cls(
            state=str(payload.get("state", "closed")),
            failure_count=int(payload.get("failure_count", 0)),
            last_failure_at=_parse_datetime(
                str(payload.get("last_failure_at"))
            )
            if payload.get("last_failure_at")
            else None,
            opened_at=_parse_datetime(str(payload.get("opened_at")))
            if payload.get("opened_at")
            else None,
            half_open_attempts=int(payload.get("half_open_attempts", 0)),
        )


@dataclass(frozen=True)
class CircuitDecision:
    allowed: bool
    state: str
    reason: str | None = None


class CircuitBreakerStore:
    def __init__(
        self,
        path: Path,
        *,
        now_provider: Callable[[], datetime] = _utc_now,
    ) -> None:
        self.path = Path(path)
        self._now = now_provider

    def load(self) -> dict[str, CircuitState]:
        if not self.path.exists():
            return {}
        raw = json.loads(self.path.read_text(encoding="utf-8"))
        raw_states = raw.get("states", raw)
        if not isinstance(raw_states, dict):
            return {}
        states: dict[str, CircuitState] = {}
        for skill_id, payload in raw_states.items():
            if isinstance(payload, dict):
                states[str(skill_id)] = CircuitState.from_dict(payload)
        return states

    def save(self, states: dict[str, CircuitState]) -> None:
        payload = {"states": {key: value.to_dict() for key, value in states.items()}}
        with file_lock(self.path):
            atomic_write_text(
                self.path,
                json.dumps(payload, ensure_ascii=True, indent=2),
                encoding="utf-8",
            )

    def now(self) -> datetime:
        return _ensure_utc(self._now())


class CircuitBreaker:
    def __init__(
        self,
        store: CircuitBreakerStore,
        config: CircuitBreakerConfig | None = None,
    ) -> None:
        self._store = store
        self._config = config or CircuitBreakerConfig()

    def allow(self, skill_id: str) -> CircuitDecision:
        states = self._store.load()
        state = states.get(skill_id, CircuitState())
        now = self._store.now()

        if state.state == "open":
            if self._open_expired(state, now):
                state = CircuitState(state="half_open")
                states[skill_id] = state
                self._store.save(states)
                return CircuitDecision(allowed=True, state=state.state)
            return CircuitDecision(
                allowed=False, state=state.state, reason="circuit_open"
            )

        if state.state == "half_open":
            if state.half_open_attempts >= self._config.half_open_max_attempts:
                return CircuitDecision(
                    allowed=False, state=state.state, reason="circuit_half_open"
                )
            state.half_open_attempts += 1
            states[skill_id] = state
            self._store.save(states)
            return CircuitDecision(allowed=True, state=state.state)

        if self._failure_window_expired(state, now):
            state.failure_count = 0
            state.last_failure_at = None
            states[skill_id] = state
            self._store.save(states)

        return CircuitDecision(allowed=True, state=state.state)

    def record_success(self, skill_id: str) -> None:
        states = self._store.load()
        states[skill_id] = CircuitState()
        self._store.save(states)

    def record_failure(self, skill_id: str) -> None:
        states = self._store.load()
        state = states.get(skill_id, CircuitState())
        now = self._store.now()

        if state.state == "open":
            return
        if state.state == "half_open":
            state = CircuitState(
                state="open",
                failure_count=self._config.failure_threshold,
                last_failure_at=now,
                opened_at=now,
            )
            states[skill_id] = state
            self._store.save(states)
            return

        if self._failure_window_expired(state, now):
            state.failure_count = 0
        state.failure_count += 1
        state.last_failure_at = now
        if state.failure_count >= self._config.failure_threshold:
            state.state = "open"
            state.opened_at = now
        states[skill_id] = state
        self._store.save(states)

    def _open_expired(self, state: CircuitState, now: datetime) -> bool:
        opened_at = state.opened_at or now
        return now >= opened_at + timedelta(seconds=self._config.open_seconds)

    def _failure_window_expired(self, state: CircuitState, now: datetime) -> bool:
        last_failure = state.last_failure_at
        if not last_failure:
            return False
        return now - last_failure > timedelta(seconds=self._config.window_seconds)


def _env_int(name: str, default: int) -> int:
    raw = os.getenv(name)
    if raw is None:
        return default
    try:
        return int(str(raw).strip())
    except ValueError:
        return default



from __future__ import annotations

import functools
import inspect
from dataclasses import dataclass
from typing import Any, Callable, Optional
from pathlib import Path

from skillos.skills.models import SkillMetadata

@dataclass
class Context:
    """Execution context passed to skill functions."""
    role: Optional[str] = None
    attributes: dict[str, object] = None
    request_id: Optional[str] = None
    
    def log(self, message: str, **kwargs):
        # Placeholder for structured logging integration
        print(f"[{self.request_id}] {message} {kwargs}")


def skill(
    name: str = None, 
    description: str = None, 
    version: str = "1.0.0",
    tags: list[str] = None,
    deprecated: bool = False,
    replacement_id: str = None,
    deprecation_reason: str = None,
) -> Callable:
    """
    Decorator to register a function as a SkillOS skill.
    
    Usage:
        @skill(name="my-skill", description="Does something useful")
        def my_skill(payload: str):
            return "Done"
            
    The decorator attaches metadata to the function which can be read by
    loaders or the registry to auto-register the skill without YAML files.
    """
    def decorator(func: Callable) -> Callable:
        # Infer name from function name if not provided
        skill_name = name or func.__name__.replace("_", "-")
        skill_desc = description or (func.__doc__ or "").strip()
        
        # Attach metadata to the function object
        func._skill_metadata = SkillMetadata(
            id=f"local/{skill_name}", # Default to local domain for decorated skills
            name=skill_name,
            description=skill_desc,
            version=version,
            entrypoint=f"{func.__module__}:{func.__name__}",
            tags=tags or [],
            deprecated=deprecated,
            replacement_id=replacement_id,
            deprecation_reason=deprecation_reason,
        )
        
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Inject context if requested
            sig = inspect.signature(func)
            if "ctx" in sig.parameters and "ctx" not in kwargs:
                # In a real execution, Orchestrator would inject this.
                # For direct calls, we inject a dummy context.
                kwargs["ctx"] = Context(request_id="direct-call")
                
            return func(*args, **kwargs)
            
        return wrapper
    return decorator



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
import json
from pathlib import Path

from skillos.storage import atomic_write_text
from skillos.tenancy import resolve_tenant_root
from skillos.storage_backend import (
    pg_connect,
    require_postgres_dsn,
    resolve_tenant_id,
    storage_backend_from_env,
)

DEFAULT_CONFIDENCE = 0.5


def default_feedback_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "feedback" / "confidence.json"


def normalize_skill_id(skill_id: str) -> str:
    if "/" in skill_id:
        return skill_id.replace("/", ".", 1)
    return skill_id


class FeedbackOutcome(str, Enum):
    POSITIVE = "positive"
    NEGATIVE = "negative"


@dataclass
class FeedbackRecord:
    confidence: float
    positive: int = 0
    negative: int = 0


class FeedbackStore:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)

    def load(self) -> dict[str, FeedbackRecord]:
        if not self.path.exists():
            return {}
        raw = json.loads(self.path.read_text(encoding="utf-8"))
        skills = raw.get("skills", raw)
        records: dict[str, FeedbackRecord] = {}
        for skill_id, payload in skills.items():
            records[skill_id] = FeedbackRecord(
                confidence=float(payload.get("confidence", DEFAULT_CONFIDENCE)),
                positive=int(payload.get("positive", 0)),
                negative=int(payload.get("negative", 0)),
            )
        return records

    def save(self, records: dict[str, FeedbackRecord]) -> None:
        payload = {
            "skills": {
                skill_id: {
                    "confidence": record.confidence,
                    "positive": record.positive,
                    "negative": record.negative,
                }
                for skill_id, record in records.items()
            }
        }
        atomic_write_text(
            self.path,
            json.dumps(payload, ensure_ascii=True),
            encoding="utf-8",
        )


class FeedbackStorePostgres:
    def __init__(self, dsn: str, *, tenant_id: str, schema: str = "skillos") -> None:
        self._dsn = dsn
        self._tenant_id = tenant_id
        self._schema = schema
        self._ensure_table()

    def load(self) -> dict[str, FeedbackRecord]:
        query = (
            f"SELECT skill_id, confidence, positive, negative "
            f"FROM {self._schema}.feedback WHERE tenant_id = %s"
        )
        with pg_connect(self._dsn) as conn:
            rows = conn.execute(query, (self._tenant_id,)).fetchall()
        records: dict[str, FeedbackRecord] = {}
        for row in rows:
            skill_id = str(row.get("skill_id") or "")
            records[skill_id] = FeedbackRecord(
                confidence=float(row.get("confidence", DEFAULT_CONFIDENCE)),
                positive=int(row.get("positive", 0)),
                negative=int(row.get("negative", 0)),
            )
        return records

    def save(self, records: dict[str, FeedbackRecord]) -> None:
        with pg_connect(self._dsn) as conn:
            for skill_id, record in records.items():
                conn.execute(
                    f"""
                    INSERT INTO {self._schema}.feedback
                        (tenant_id, skill_id, confidence, positive, negative)
                    VALUES (%s, %s, %s, %s, %s)
                    ON CONFLICT (tenant_id, skill_id) DO UPDATE SET
                        confidence = EXCLUDED.confidence,
                        positive = EXCLUDED.positive,
                        negative = EXCLUDED.negative,
                        updated_at = now()
                    """,
                    (
                        self._tenant_id,
                        skill_id,
                        record.confidence,
                        record.positive,
                        record.negative,
                    ),
                )

    def _ensure_table(self) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {self._schema}")
            conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self._schema}.feedback (
                    tenant_id TEXT NOT NULL,
                    skill_id TEXT NOT NULL,
                    confidence DOUBLE PRECISION NOT NULL,
                    positive INTEGER NOT NULL,
                    negative INTEGER NOT NULL,
                    updated_at TIMESTAMPTZ NOT NULL DEFAULT now(),
                    PRIMARY KEY (tenant_id, skill_id)
                )
                """
            )


def feedback_store_from_env(root: Path) -> FeedbackStore | FeedbackStorePostgres:
    config = storage_backend_from_env()
    if config.backend != "postgres":
        return FeedbackStore(default_feedback_path(root))
    dsn = require_postgres_dsn(config, context="feedback")
    tenant_id = resolve_tenant_id(Path(root))
    return FeedbackStorePostgres(
        dsn,
        tenant_id=tenant_id,
        schema=config.postgres_schema,
    )


class FeedbackTracker:
    def __init__(self, store: FeedbackStore, default_confidence: float = DEFAULT_CONFIDENCE) -> None:
        self._store = store
        self._default_confidence = default_confidence

    def get_confidence(self, skill_id: str) -> float:
        normalized = normalize_skill_id(skill_id)
        record = self._store.load().get(normalized)
        if record:
            return record.confidence
        return self._default_confidence

    def record_feedback(self, skill_id: str, outcome: FeedbackOutcome) -> FeedbackRecord:
        normalized = normalize_skill_id(skill_id)
        records = self._store.load()
        record = records.get(
            normalized,
            FeedbackRecord(confidence=self._default_confidence),
        )
        if outcome == FeedbackOutcome.POSITIVE:
            record.positive += 1
            record.confidence = record.confidence + (1 - record.confidence) * 0.1
        else:
            record.negative += 1
            record.confidence = record.confidence * 0.9
        records[normalized] = record
        self._store.save(records)
        return record

    def record_correction(
        self, selected_skill_id: str, expected_skill_id: str | None
    ) -> dict[str, FeedbackRecord]:
        normalized_selected = normalize_skill_id(selected_skill_id)
        normalized_expected = (
            normalize_skill_id(expected_skill_id)
            if expected_skill_id
            else normalized_selected
        )
        updates: dict[str, FeedbackRecord] = {}
        if normalized_selected == normalized_expected:
            updates[normalized_selected] = self.record_feedback(
                normalized_selected, FeedbackOutcome.POSITIVE
            )
        else:
            updates[normalized_selected] = self.record_feedback(
                normalized_selected, FeedbackOutcome.NEGATIVE
            )
            updates[normalized_expected] = self.record_feedback(
                normalized_expected, FeedbackOutcome.POSITIVE
            )
        return updates



from __future__ import annotations

from datetime import datetime, timezone
import json
from pathlib import Path
import time
from typing import Iterable

from skillos.budget import budget_config_from_env
from skillos.routing import SkillRouter
from skillos.telemetry import token_count


def load_golden_queries(path: Path) -> list[dict[str, object]]:
    return json.loads(Path(path).read_text(encoding="utf-8"))


def _percentile(values: list[float], percentile: float) -> float:
    if not values:
        return 0.0
    sorted_values = sorted(values)
    index = int(round((percentile / 100) * (len(sorted_values) - 1)))
    return sorted_values[index]


def _mean(values: Iterable[float]) -> float:
    values_list = list(values)
    if not values_list:
        return 0.0
    return sum(values_list) / len(values_list)


def build_metrics_summary(router: SkillRouter, golden_queries: list[dict[str, object]]) -> str:
    total_requests = len(golden_queries)
    if total_requests == 0:
        payload = {
            "generated_at": datetime.now(timezone.utc).isoformat(),
            "total_requests": 0,
            "metrics": {},
        }
        return json.dumps(payload, ensure_ascii=True, indent=2)

    budget_config = budget_config_from_env()

    correct = 0
    top3 = 0
    no_skill_found = 0
    override_rate = 0.0

    routing_latencies: list[float] = []
    request_latencies: list[float] = []
    confidences: list[float] = []
    outcomes: list[float] = []

    tokens_total = 0
    cost_total = 0.0
    skill_totals: dict[str, int] = {}
    skill_success: dict[str, int] = {}
    error_counts: dict[str, int] = {}

    for item in golden_queries:
        query = str(item["query"])
        expected_skill = str(item["expected_skill_id"])
        tokens_total += token_count(query)
        cost_total += budget_config.standard_cost

        start = time.perf_counter()
        result = router.route(query)
        routing_latency = (time.perf_counter() - start) * 1000
        routing_latencies.append(routing_latency)
        request_latencies.append(routing_latency)

        predicted_skill = result.skill_id
        is_correct = predicted_skill == expected_skill
        if is_correct:
            correct += 1
        top3_candidates = [candidate.skill_id for candidate in result.candidates[:3]]
        if expected_skill in top3_candidates:
            top3 += 1
        if result.status == "no_skill_found":
            no_skill_found += 1

        confidences.append(result.confidence)
        outcomes.append(1.0 if is_correct else 0.0)

        skill_totals[expected_skill] = skill_totals.get(expected_skill, 0) + 1
        if is_correct:
            skill_success[expected_skill] = skill_success.get(expected_skill, 0) + 1

        error_class = None
        if result.status == "no_skill_found":
            error_class = "no_skill_found"
        elif not is_correct:
            error_class = "misroute"
        if error_class:
            error_counts[error_class] = error_counts.get(error_class, 0) + 1

    skill_success_rate = {
        skill_id: skill_success.get(skill_id, 0) / total
        for skill_id, total in skill_totals.items()
    }
    error_rate_by_class = {
        error_class: count / total_requests
        for error_class, count in error_counts.items()
    }

    brier_scores = [
        (confidence - outcome) ** 2
        for confidence, outcome in zip(confidences, outcomes)
    ]

    payload = {
        "generated_at": datetime.now(timezone.utc).isoformat(),
        "total_requests": total_requests,
        "metrics": {
            "routing_top1_accuracy": correct / total_requests,
            "routing_top3_accuracy": top3 / total_requests,
            "routing_no_skill_found_rate": no_skill_found / total_requests,
            "routing_override_rate": override_rate,
            "routing_confidence_brier": _mean(brier_scores),
            "routing_p95_ms": _percentile(routing_latencies, 95),
            "request_p95_ms": _percentile(request_latencies, 95),
            "skill_success_rate": skill_success_rate,
            "error_rate_by_class": error_rate_by_class,
            "tokens_per_request": tokens_total / total_requests,
            "cost_per_request": cost_total / total_requests,
        },
    }
    return json.dumps(payload, ensure_ascii=True, indent=2)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
import sqlite3
import time
from typing import Callable, Iterator
from uuid import uuid4
from contextlib import contextmanager

from skillos.orchestrator import Orchestrator
from skillos.execution_planner import build_execution_plan
from skillos.routing import to_internal_id, to_public_id
from skillos.telemetry import (
    EventLogger,
    default_log_path,
    log_job_failed,
    log_job_started,
    log_job_succeeded,
    new_request_id,
)
from skillos.tenancy import resolve_tenant_root
from skillos.storage_backend import (
    pg_connect,
    require_postgres_dsn,
    resolve_tenant_id,
    storage_backend_from_env,
)


def default_jobs_db_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "runtime" / "jobs.db"


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _ensure_utc(value: datetime) -> datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=timezone.utc)
    return value.astimezone(timezone.utc)


def _format_datetime(value: datetime | None) -> str | None:
    if value is None:
        return None
    return _ensure_utc(value).isoformat()


def _parse_datetime(value: str | None) -> datetime | None:
    if not value:
        return None
    parsed = datetime.fromisoformat(value)
    return _ensure_utc(parsed)


def retry_backoff(
    retry_count: int,
    *,
    base_seconds: int = 5,
    max_seconds: int = 300,
) -> timedelta:
    if retry_count <= 0:
        return timedelta(0)
    delay = base_seconds * (2 ** (retry_count - 1))
    return timedelta(seconds=min(max_seconds, delay))


@dataclass
class JobRecord:
    job_id: str
    skill_id: str
    payload: str
    status: str
    retries: int
    max_retries: int
    created_at: datetime
    updated_at: datetime
    next_run_at: datetime | None = None
    last_error: str | None = None

    def is_due(self, now: datetime) -> bool:
        if self.status != "queued":
            return False
        due_at = self.next_run_at or self.created_at
        return _ensure_utc(due_at) <= _ensure_utc(now)

    def mark_running(self, now: datetime) -> None:
        self.status = "running"
        self.updated_at = _ensure_utc(now)
        self.next_run_at = None

    def mark_succeeded(self, now: datetime) -> None:
        self.status = "succeeded"
        self.updated_at = _ensure_utc(now)
        self.next_run_at = None
        self.last_error = None

    def mark_failed(
        self,
        now: datetime,
        error: str,
        *,
        base_seconds: int = 5,
        max_seconds: int = 300,
    ) -> bool:
        self.retries += 1
        self.last_error = error
        self.updated_at = _ensure_utc(now)
        if self.retries <= self.max_retries:
            delay = retry_backoff(
                self.retries,
                base_seconds=base_seconds,
                max_seconds=max_seconds,
            )
            self.status = "queued"
            self.next_run_at = self.updated_at + delay
            return True
        self.status = "failed"
        self.next_run_at = None
        return False

    def to_dict(self) -> dict[str, object]:
        return {
            "job_id": self.job_id,
            "skill_id": self.skill_id,
            "payload": self.payload,
            "status": self.status,
            "retries": self.retries,
            "max_retries": self.max_retries,
            "created_at": _format_datetime(self.created_at),
            "updated_at": _format_datetime(self.updated_at),
            "next_run_at": _format_datetime(self.next_run_at),
            "last_error": self.last_error,
        }

    @classmethod
    def from_row(cls, row: sqlite3.Row) -> "JobRecord":
        return cls(
            job_id=str(row["job_id"]),
            skill_id=str(row["skill_id"]),
            payload=str(row["payload"]),
            status=str(row["status"]),
            retries=int(row["retries"]),
            max_retries=int(row["max_retries"]),
            created_at=_parse_datetime(row["created_at"]) or _utc_now(),
            updated_at=_parse_datetime(row["updated_at"]) or _utc_now(),
            next_run_at=_parse_datetime(row["next_run_at"]),
            last_error=str(row["last_error"]) if row["last_error"] else None,
        )


class JobStore:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)
        self._ensure_db()

    def enqueue(
        self,
        skill_id: str,
        *,
        payload: str = "ok",
        max_retries: int = 0,
        job_id: str | None = None,
        now: datetime | None = None,
    ) -> JobRecord:
        created_at = _ensure_utc(now or _utc_now())
        record = JobRecord(
            job_id=job_id or uuid4().hex,
            skill_id=skill_id,
            payload=payload,
            status="queued",
            retries=0,
            max_retries=max_retries,
            created_at=created_at,
            updated_at=created_at,
            next_run_at=created_at,
            last_error=None,
        )
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO jobs (
                    job_id, skill_id, payload, status, retries, max_retries,
                    created_at, updated_at, next_run_at, last_error
                ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    record.job_id,
                    record.skill_id,
                    record.payload,
                    record.status,
                    record.retries,
                    record.max_retries,
                    _format_datetime(record.created_at),
                    _format_datetime(record.updated_at),
                    _format_datetime(record.next_run_at),
                    record.last_error,
                ),
            )
        return record

    def fetch_due(
        self,
        *,
        now: datetime | None = None,
        limit: int | None = None,
    ) -> list[JobRecord]:
        due_at = _format_datetime(_ensure_utc(now or _utc_now()))
        query = (
            "SELECT * FROM jobs WHERE status = ? "
            "AND (next_run_at IS NULL OR next_run_at <= ?) "
            "ORDER BY created_at"
        )
        params: list[object] = ["queued", due_at]
        if limit is not None:
            query += " LIMIT ?"
            params.append(int(limit))
        with self._connect() as conn:
            rows = conn.execute(query, params).fetchall()
        return [JobRecord.from_row(row) for row in rows]

    def get(self, job_id: str) -> JobRecord | None:
        with self._connect() as conn:
            row = conn.execute(
                "SELECT * FROM jobs WHERE job_id = ?",
                (job_id,),
            ).fetchone()
        return JobRecord.from_row(row) if row else None

    def save(self, record: JobRecord) -> None:
        with self._connect() as conn:
            conn.execute(
                """
                UPDATE jobs
                SET skill_id = ?,
                    payload = ?,
                    status = ?,
                    retries = ?,
                    max_retries = ?,
                    created_at = ?,
                    updated_at = ?,
                    next_run_at = ?,
                    last_error = ?
                WHERE job_id = ?
                """,
                (
                    record.skill_id,
                    record.payload,
                    record.status,
                    record.retries,
                    record.max_retries,
                    _format_datetime(record.created_at),
                    _format_datetime(record.updated_at),
                    _format_datetime(record.next_run_at),
                    record.last_error,
                    record.job_id,
                ),
            )

    def list_all(self) -> list[JobRecord]:
        with self._connect() as conn:
            rows = conn.execute("SELECT * FROM jobs ORDER BY created_at").fetchall()
        return [JobRecord.from_row(row) for row in rows]

    def requeue_dead_letters(self, max_retries_bump: int = 1) -> int:
        """
        Resets 'failed' jobs to 'queued' and increments max_retries to allow another attempt.
        Returns count of requeued jobs.
        """
        with self._connect() as conn:
            # Atomic update for SQLite
            cursor = conn.execute(
                """
                UPDATE jobs
                SET status = 'queued',
                    max_retries = max_retries + ?,
                    next_run_at = ?,
                    updated_at = ?
                WHERE status = 'failed'
                """,
                (max_retries_bump, _format_datetime(_utc_now()), _format_datetime(_utc_now()))
            )
            return cursor.rowcount

    def _ensure_db(self) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        conn = sqlite3.connect(self.path)
        try:
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS jobs (
                    job_id TEXT PRIMARY KEY,
                    skill_id TEXT NOT NULL,
                    payload TEXT NOT NULL,
                    status TEXT NOT NULL,
                    retries INTEGER NOT NULL,
                    max_retries INTEGER NOT NULL,
                    created_at TEXT NOT NULL,
                    updated_at TEXT NOT NULL,
                    next_run_at TEXT,
                    last_error TEXT
                )
                """
            )
            conn.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_jobs_status_run
                ON jobs (status, next_run_at)
                """
            )
            conn.commit()
        finally:
            conn.close()


    @contextmanager
    def _connect(self) -> Iterator[sqlite3.Connection]:
        conn = sqlite3.connect(self.path)
        conn.row_factory = sqlite3.Row
        try:
            yield conn
            conn.commit()
        except:
            conn.rollback()
            raise
        finally:
            conn.close()


class JobStorePostgres:
    def __init__(self, dsn: str, *, tenant_id: str, schema: str = "skillos") -> None:
        self._dsn = dsn
        self._tenant_id = tenant_id
        self._schema = schema
        self._ensure_db()

    def enqueue(
        self,
        skill_id: str,
        *,
        payload: str = "ok",
        max_retries: int = 0,
        job_id: str | None = None,
        now: datetime | None = None,
    ) -> JobRecord:
        created_at = _ensure_utc(now or _utc_now())
        record = JobRecord(
            job_id=job_id or uuid4().hex,
            skill_id=skill_id,
            payload=payload,
            status="queued",
            retries=0,
            max_retries=max_retries,
            created_at=created_at,
            updated_at=created_at,
            next_run_at=created_at,
            last_error=None,
        )
        with pg_connect(self._dsn) as conn:
            conn.execute(
                f"""
                INSERT INTO {self._schema}.jobs (
                    tenant_id, job_id, skill_id, payload, status, retries, max_retries,
                    created_at, updated_at, next_run_at, last_error
                ) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
                """,
                (
                    self._tenant_id,
                    record.job_id,
                    record.skill_id,
                    record.payload,
                    record.status,
                    record.retries,
                    record.max_retries,
                    _format_datetime(record.created_at),
                    _format_datetime(record.updated_at),
                    _format_datetime(record.next_run_at),
                    record.last_error,
                ),
            )
        return record

    def fetch_due(
        self,
        *,
        now: datetime | None = None,
        limit: int | None = None,
    ) -> list[JobRecord]:
        due_at = _format_datetime(_ensure_utc(now or _utc_now()))
        query = (
            f"SELECT * FROM {self._schema}.jobs "
            "WHERE tenant_id = %s AND status = %s "
            "AND (next_run_at IS NULL OR next_run_at <= %s) "
            "ORDER BY created_at"
        )
        params: list[object] = [self._tenant_id, "queued", due_at]
        if limit is not None:
            query += " LIMIT %s"
            params.append(int(limit))
        with pg_connect(self._dsn) as conn:
            rows = conn.execute(query, params).fetchall()
        return [JobRecord.from_row(_pg_row_to_sqlite(row)) for row in rows]

    def get(self, job_id: str) -> JobRecord | None:
        with pg_connect(self._dsn) as conn:
            row = conn.execute(
                f"SELECT * FROM {self._schema}.jobs WHERE tenant_id = %s AND job_id = %s",
                (self._tenant_id, job_id),
            ).fetchone()
        return JobRecord.from_row(_pg_row_to_sqlite(row)) if row else None

    def save(self, record: JobRecord) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(
                f"""
                UPDATE {self._schema}.jobs
                SET skill_id = %s,
                    payload = %s,
                    status = %s,
                    retries = %s,
                    max_retries = %s,
                    created_at = %s,
                    updated_at = %s,
                    next_run_at = %s,
                    last_error = %s
                WHERE tenant_id = %s AND job_id = %s
                """,
                (
                    record.skill_id,
                    record.payload,
                    record.status,
                    record.retries,
                    record.max_retries,
                    _format_datetime(record.created_at),
                    _format_datetime(record.updated_at),
                    _format_datetime(record.next_run_at),
                    record.last_error,
                    self._tenant_id,
                    record.job_id,
                ),
            )

    def list_all(self) -> list[JobRecord]:
        with pg_connect(self._dsn) as conn:
            rows = conn.execute(
                f"SELECT * FROM {self._schema}.jobs "
                "WHERE tenant_id = %s ORDER BY created_at",
                (self._tenant_id,),
            ).fetchall()
        return [JobRecord.from_row(_pg_row_to_sqlite(row)) for row in rows]

    def requeue_dead_letters(self, max_retries_bump: int = 1) -> int:
        with pg_connect(self._dsn) as conn:
            cursor = conn.execute(
                f"""
                UPDATE {self._schema}.jobs
                SET status = %s,
                    max_retries = max_retries + %s,
                    next_run_at = %s,
                    updated_at = %s
                WHERE tenant_id = %s AND status = %s
                """,
                (
                    "queued",
                    max_retries_bump,
                    _format_datetime(_utc_now()),
                    _format_datetime(_utc_now()),
                    self._tenant_id,
                    "failed",
                )
            )
            return cursor.rowcount

    def _ensure_db(self) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {self._schema}")
            conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self._schema}.jobs (
                    tenant_id TEXT NOT NULL,
                    job_id TEXT NOT NULL,
                    skill_id TEXT NOT NULL,
                    payload TEXT NOT NULL,
                    status TEXT NOT NULL,
                    retries INTEGER NOT NULL,
                    max_retries INTEGER NOT NULL,
                    created_at TEXT NOT NULL,
                    updated_at TEXT NOT NULL,
                    next_run_at TEXT,
                    last_error TEXT,
                    PRIMARY KEY (tenant_id, job_id)
                )
                """
            )
            conn.execute(
                f"""
                CREATE INDEX IF NOT EXISTS jobs_status_run_idx
                ON {self._schema}.jobs (tenant_id, status, next_run_at)
                """
            )


def _pg_row_to_sqlite(row: dict[str, object]) -> sqlite3.Row:
    class _Row(dict):
        def __getattr__(self, name: str):
            return self[name]

    return _Row(row)  # type: ignore[return-value]


def job_store_from_env(root: Path) -> JobStore | JobStorePostgres:
    config = storage_backend_from_env()
    if config.backend != "postgres":
        return JobStore(default_jobs_db_path(root))
    dsn = require_postgres_dsn(config, context="jobs")
    tenant_id = resolve_tenant_id(Path(root))
    return JobStorePostgres(
        dsn,
        tenant_id=tenant_id,
        schema=config.postgres_schema,
    )


class JobWorker:
    def __init__(
        self,
        root_path: Path,
        store: JobStore,
        *,
        log_path: Path | None = None,
        now_provider: Callable[[], datetime] = _utc_now,
    ) -> None:
        self.root_path = Path(root_path)
        self.store = store
        self.log_path = log_path or default_log_path(self.root_path)
        self._now = now_provider

        self.orchestrator = Orchestrator(self.root_path, self.log_path)

    def run_once(self, *, limit: int | None = None) -> list[JobRecord]:
        now = self._now()
        jobs = self.store.fetch_due(now=now, limit=limit)
        results: list[JobRecord] = []
        for job in jobs:
            results.append(self._execute_job(job))
        return results

    def _execute_job(self, job: JobRecord) -> JobRecord:
        now = self._now()
        logger = EventLogger(self.log_path, request_id=new_request_id())
        job.mark_running(now)
        self.store.save(job)
        log_job_started(
            logger,
            job_id=job.job_id,
            skill_id=job.skill_id,
            status=job.status,
            retries=job.retries,
            max_retries=job.max_retries,
        )

        start = time.perf_counter()
        internal_skill_id = to_internal_id(job.skill_id)
        metadata = self.orchestrator.registry.get(internal_skill_id)
        if metadata is None:
            return self._fail_job(job, logger, now, "unknown_skill", start)

        plan = build_execution_plan(
            to_public_id(internal_skill_id),
            internal_skill_id,
            job.payload,
        )

        try:
            result = self.orchestrator.execute_plan(
                plan,
                execute=True,
                dry_run=False,
                approval=None,
                approval_token=None,
                role=None,
                attributes=None,
                plan_path=None,
                debug_trace=None,
                logger=logger,
                request_start=start,
                metadata=metadata,
                warnings=None,
                skill_selected=None,
            )
        except Exception as exc:  # pragma: no cover - execution exceptions
            return self._fail_job(
                job,
                logger,
                now,
                exc.__class__.__name__,
                start,
                log_execution_result=False,
            )

        if result.get("status") != "success":
            reason = result.get("reason") or result.get("policy_id") or "blocked"
            return self._fail_job(
                job,
                logger,
                now,
                reason,
                start,
                log_execution_result=False,
            )

        duration_ms = (time.perf_counter() - start) * 1000
        job.mark_succeeded(now)
        self.store.save(job)
        log_job_succeeded(
            logger,
            job_id=job.job_id,
            skill_id=job.skill_id,
            status=job.status,
            duration_ms=duration_ms,
        )
        return job

    def _fail_job(
        self,
        job: JobRecord,
        logger: EventLogger,
        now: datetime,
        error_class: str,
        start: float,
        *,
        log_execution_result: bool = True,
    ) -> JobRecord:
        duration_ms = (time.perf_counter() - start) * 1000
        if log_execution_result:
            logger.log(
                "execution_result",
                status="error",
                duration_ms=duration_ms,
                error_class=error_class,
            )
        will_retry = job.mark_failed(now, error_class)
        self.store.save(job)
        log_job_failed(
            logger,
            job_id=job.job_id,
            skill_id=job.skill_id,
            status=job.status,
            error_class=error_class,
            retries=job.retries,
            max_retries=job.max_retries,
            next_run_at=_format_datetime(job.next_run_at),
            retrying=will_retry,
        )
        return job



from __future__ import annotations

from dataclasses import dataclass
import base64
import hashlib
import hmac
import json
import os
import re
import time
from typing import Mapping


class JwtValidationError(ValueError):
    pass


_JWT_TOKEN_ENV = "SKILLOS_JWT_TOKEN"
_ALG_PATTERN = re.compile(r"^HS(256|384|512)$")
_DEFAULT_CLOCK_SKEW = 60


@dataclass(frozen=True)
class JwtConfig:
    secret: str | None = None
    issuer: str | None = None
    audience: str | None = None
    algorithm: str = "HS256"
    clock_skew_seconds: int = _DEFAULT_CLOCK_SKEW
    require_exp: bool = True
    allow_unverified: bool = False


@dataclass(frozen=True)
class JwtClaims:
    subject: str | None
    tenant_id: str | None
    role: str | None
    permissions: list[str]
    claims: dict[str, object]
    verified: bool

    @property
    def attributes(self) -> dict[str, object]:
        payload = {key: value for key, value in self.claims.items() if key != "role"}
        if self.subject is not None:
            payload.setdefault("subject", self.subject)
        if self.tenant_id is not None:
            payload.setdefault("tenant_id", self.tenant_id)
        if self.permissions:
            payload.setdefault("permissions", list(self.permissions))
        payload.setdefault("verified", self.verified)
        return payload


def jwt_config_from_env(env: Mapping[str, str] | None = None) -> JwtConfig:
    env_map = env or os.environ
    return JwtConfig(
        secret=_env_value(env_map, "SKILLOS_JWT_SECRET"),
        issuer=_env_value(env_map, "SKILLOS_JWT_ISSUER"),
        audience=_env_value(env_map, "SKILLOS_JWT_AUDIENCE"),
        algorithm=_env_value(env_map, "SKILLOS_JWT_ALGORITHM") or "HS256",
        clock_skew_seconds=_env_int(
            env_map,
            "SKILLOS_JWT_CLOCK_SKEW_SECONDS",
            _DEFAULT_CLOCK_SKEW,
        ),
        require_exp=_env_bool(env_map, "SKILLOS_JWT_REQUIRE_EXP", True),
        allow_unverified=_resolve_allow_unverified(env_map),
    )


def _resolve_allow_unverified(env_map: Mapping[str, str]) -> bool:
    is_prod = _env_value(env_map, "SKILLOS_ENV") in ("prod", "production")
    requested = _env_bool(env_map, "SKILLOS_JWT_ALLOW_UNVERIFIED", False)
    if is_prod and requested:
        # Log warning here in real app, but for now just enforce security
        return False
    return requested


def jwt_token_from_env(env: Mapping[str, str] | None = None) -> str | None:
    env_map = env or os.environ
    raw = env_map.get(_JWT_TOKEN_ENV)
    if not raw:
        return None
    token = raw.strip()
    return token or None


def decode_jwt(token: str, config: JwtConfig | None = None) -> JwtClaims:
    if not token or not isinstance(token, str):
        raise JwtValidationError("jwt_missing")
    segments = token.strip().split(".")
    if len(segments) != 3:
        raise JwtValidationError("jwt_invalid_format")

    header = _decode_segment(segments[0], "header")
    payload = _decode_segment(segments[1], "payload")
    signature = _decode_bytes(segments[2], "signature")

    if not isinstance(header, dict) or not isinstance(payload, dict):
        raise JwtValidationError("jwt_invalid_payload")

    config = config or jwt_config_from_env()
    algorithm = str(header.get("alg") or "").upper()
    if not algorithm or algorithm == "NONE":
        raise JwtValidationError("jwt_alg_missing")
    if not _ALG_PATTERN.match(algorithm):
        raise JwtValidationError("jwt_alg_unsupported")
    if algorithm != config.algorithm.upper():
        raise JwtValidationError("jwt_alg_mismatch")

    verified = _verify_signature(
        segments[0],
        segments[1],
        signature,
        algorithm,
        config,
    )
    _validate_claims(payload, config)

    subject = _extract_string(payload.get("sub"))
    tenant_id = _extract_tenant_id(payload)
    role = _extract_role(payload)
    permissions = _extract_permissions(payload)

    return JwtClaims(
        subject=subject,
        tenant_id=tenant_id,
        role=role,
        permissions=permissions,
        claims=dict(payload),
        verified=verified,
    )


def _verify_signature(
    header_segment: str,
    payload_segment: str,
    signature: bytes,
    algorithm: str,
    config: JwtConfig,
) -> bool:
    secret = (config.secret or "").strip()
    if not secret:
        if config.allow_unverified:
            return False
        raise JwtValidationError("jwt_secret_missing")

    signing_input = f"{header_segment}.{payload_segment}".encode("ascii")
    digest = _hash_for_alg(algorithm)
    expected = hmac.new(secret.encode("utf-8"), signing_input, digest).digest()
    if not hmac.compare_digest(expected, signature):
        raise JwtValidationError("jwt_signature_invalid")
    return True


def _hash_for_alg(algorithm: str):
    if algorithm == "HS256":
        return hashlib.sha256
    if algorithm == "HS384":
        return hashlib.sha384
    if algorithm == "HS512":
        return hashlib.sha512
    raise JwtValidationError("jwt_alg_unsupported")


def _validate_claims(payload: dict[str, object], config: JwtConfig) -> None:
    now = time.time()
    skew = max(0, int(config.clock_skew_seconds))

    exp = payload.get("exp")
    if exp is None:
        if config.require_exp:
            raise JwtValidationError("jwt_exp_missing")
    else:
        exp_ts = _as_timestamp(exp, "exp")
        if now > exp_ts + skew:
            raise JwtValidationError("jwt_expired")

    nbf = payload.get("nbf")
    if nbf is not None:
        nbf_ts = _as_timestamp(nbf, "nbf")
        if now + skew < nbf_ts:
            raise JwtValidationError("jwt_not_active")

    if config.issuer:
        issuer = _extract_string(payload.get("iss"))
        if issuer != config.issuer:
            raise JwtValidationError("jwt_issuer_invalid")

    if config.audience:
        audience = payload.get("aud")
        if isinstance(audience, str):
            allowed = audience == config.audience
        elif isinstance(audience, list):
            allowed = config.audience in [str(item) for item in audience]
        else:
            allowed = False
        if not allowed:
            raise JwtValidationError("jwt_audience_invalid")


def _as_timestamp(value: object, field: str) -> float:
    if isinstance(value, bool):
        raise JwtValidationError(f"jwt_{field}_invalid")
    if isinstance(value, (int, float)):
        return float(value)
    if isinstance(value, str):
        try:
            return float(value)
        except ValueError as exc:
            raise JwtValidationError(f"jwt_{field}_invalid") from exc
    raise JwtValidationError(f"jwt_{field}_invalid")


def _extract_tenant_id(payload: dict[str, object]) -> str | None:
    for key in ("tenant_id", "tenant", "tid"):
        value = _extract_string(payload.get(key))
        if value:
            return value
    return None


def _extract_role(payload: dict[str, object]) -> str | None:
    value = _extract_string(payload.get("role"))
    if value:
        return value
    raw_roles = payload.get("roles")
    if isinstance(raw_roles, (list, tuple)):
        for item in raw_roles:
            role = _extract_string(item)
            if role:
                return role
    return None


def _extract_permissions(payload: dict[str, object]) -> list[str]:
    permissions = _normalize_string_list(payload.get("permissions"))
    if permissions:
        return permissions
    scopes = _normalize_string_list(payload.get("scopes"))
    if scopes:
        return scopes
    scope = _extract_string(payload.get("scope"))
    if not scope:
        return []
    parts = [part.strip() for part in re.split(r"[\\s,]+", scope) if part.strip()]
    return parts


def _normalize_string_list(value: object) -> list[str]:
    if isinstance(value, str):
        return [item for item in [value.strip()] if item]
    if isinstance(value, (list, tuple)):
        return [item for item in (_extract_string(item) for item in value) if item]
    return []


def _extract_string(value: object) -> str | None:
    if value is None:
        return None
    text = str(value).strip()
    return text or None


def _decode_segment(segment: str, name: str) -> dict[str, object]:
    raw = _decode_bytes(segment, name)
    try:
        payload = json.loads(raw.decode("utf-8"))
    except (UnicodeDecodeError, json.JSONDecodeError) as exc:
        raise JwtValidationError(f"jwt_{name}_invalid") from exc
    if not isinstance(payload, dict):
        raise JwtValidationError(f"jwt_{name}_invalid")
    return payload


def _decode_bytes(segment: str, name: str) -> bytes:
    if not isinstance(segment, str) or not segment:
        raise JwtValidationError(f"jwt_{name}_invalid")
    padded = segment + "=" * (-len(segment) % 4)
    try:
        return base64.urlsafe_b64decode(padded.encode("ascii"))
    except (ValueError, UnicodeEncodeError) as exc:
        raise JwtValidationError(f"jwt_{name}_invalid") from exc


def _env_value(env_map: Mapping[str, str], key: str) -> str | None:
    raw = env_map.get(key)
    if raw is None:
        return None
    value = str(raw).strip()
    return value or None


def _env_bool(env_map: Mapping[str, str], key: str, default: bool) -> bool:
    raw = env_map.get(key)
    if raw is None:
        return default
    normalized = str(raw).strip().lower()
    if normalized in {"1", "true", "yes", "y"}:
        return True
    if normalized in {"0", "false", "no", "n"}:
        return False
    return default


def _env_int(env_map: Mapping[str, str], key: str, default: int) -> int:
    raw = env_map.get(key)
    if raw is None:
        return default
    try:
        return int(str(raw).strip())
    except ValueError:
        return default



import random

def run(payload: str = "ok") -> str:
    coffees = ["Cappuccino", "Espresso", "Latte", "Americano", "Flat White"]
    choice = random.choice(coffees)
    return f"Time for a {choice}! Go get one."



from __future__ import annotations

from pathlib import Path

import yaml

from skillos.routing import to_internal_id
from skillos.skills.errors import SkillValidationError
from skillos.skills.models import SkillMetadata
from skillos.skills.registry import SkillRegistry


def deprecate_skill(
    root: Path,
    skill_id: str,
    *,
    reason: str | None = None,
    replacement_id: str | None = None,
) -> Path:
    reason = _normalize_optional(reason)
    replacement_id = _normalize_optional(replacement_id)
    if not reason and not replacement_id:
        raise ValueError("deprecation_reason_or_replacement_required")

    raw, metadata_path = _load_metadata(root, skill_id)
    raw["deprecated"] = True
    if reason is not None:
        raw["deprecation_reason"] = reason
    if replacement_id is not None:
        raw["replacement_id"] = replacement_id

    _validate_metadata(raw)
    _write_metadata(metadata_path, raw)
    return metadata_path


def undeprecate_skill(root: Path, skill_id: str) -> Path:
    raw, metadata_path = _load_metadata(root, skill_id)
    raw["deprecated"] = False
    raw.pop("deprecation_reason", None)
    raw.pop("replacement_id", None)

    _validate_metadata(raw)
    _write_metadata(metadata_path, raw)
    return metadata_path


def _load_metadata(root: Path, skill_id: str) -> tuple[dict[str, object], Path]:
    registry = SkillRegistry(root)
    records = registry.load_all()
    internal_id = to_internal_id(skill_id.strip())
    record = records.get(internal_id)
    if not record:
        raise FileNotFoundError(f"Unknown skill: {skill_id}")

    metadata_path = record.source
    raw = yaml.safe_load(metadata_path.read_text(encoding="utf-8"))
    if not isinstance(raw, dict):
        raise SkillValidationError("Skill YAML must be a mapping")
    return raw, metadata_path


def _validate_metadata(raw: dict[str, object]) -> None:
    try:
        SkillMetadata.model_validate(raw)
    except Exception as exc:
        raise SkillValidationError(str(exc)) from exc


def _write_metadata(path: Path, payload: dict[str, object]) -> None:
    path.write_text(
        yaml.safe_dump(payload, sort_keys=False),
        encoding="utf-8",
    )


def _normalize_optional(value: str | None) -> str | None:
    if value is None:
        return None
    cleaned = value.strip()
    return cleaned or None


def build_deprecation_warning(
    skill_id: str,
    metadata: SkillMetadata,
) -> dict[str, object]:
    message = f"Skill {skill_id} is deprecated"
    if metadata.replacement_id:
        message = f"{message}; use {metadata.replacement_id}"
    warning: dict[str, object] = {
        "code": "deprecated_skill",
        "skill_id": skill_id,
        "message": message,
    }
    if metadata.deprecation_reason:
        warning["deprecation_reason"] = metadata.deprecation_reason
    if metadata.replacement_id:
        warning["replacement_id"] = metadata.replacement_id
    return warning



from __future__ import annotations

import os
from pathlib import Path
import re
from typing import Mapping


_TENANT_ENV = "SKILLOS_TENANT_ID"
_TENANT_DIR = "tenants"
_TENANT_RE = re.compile(r"[^A-Za-z0-9_-]+")


def normalize_tenant_id(value: str | None) -> str | None:
    if value is None:
        return None
    cleaned = str(value).strip()
    if not cleaned:
        return None
    normalized = _TENANT_RE.sub("-", cleaned)
    normalized = normalized.strip("-_")
    normalized = re.sub(r"-{2,}", "-", normalized)
    return normalized or None


def tenant_id_from_env(env: Mapping[str, str] | None = None) -> str | None:
    env_map = env or os.environ
    return normalize_tenant_id(env_map.get(_TENANT_ENV))


def resolve_tenant_root(root: Path | str, tenant_id: str | None = None) -> Path:
    root_path = Path(root)
    if tenant_id is not None:
        resolved = normalize_tenant_id(tenant_id)
    else:
        resolved = tenant_id_from_env()
    if not resolved:
        return root_path
    if _has_tenant_segment(root_path, resolved):
        return root_path
    if root_path.parts and root_path.parts[-1] == _TENANT_DIR:
        return root_path / resolved
    return root_path / _TENANT_DIR / resolved


def _has_tenant_segment(root: Path, tenant_id: str) -> bool:
    parts = root.parts
    for index in range(len(parts) - 1):
        if parts[index] == _TENANT_DIR and parts[index + 1] == tenant_id:
            return True
    return False


def tenant_id_from_path(root: Path | str) -> str | None:
    root_path = Path(root)
    parts = root_path.parts
    for index in range(len(parts) - 1):
        if parts[index] == _TENANT_DIR:
            candidate = normalize_tenant_id(parts[index + 1])
            return candidate
    return None



from __future__ import annotations

from dataclasses import dataclass, field
import hashlib
import json
from pathlib import Path

from skillos.storage import atomic_write_text
from skillos.tenancy import resolve_tenant_root
from skillos.storage_backend import (
    pg_connect,
    require_postgres_dsn,
    resolve_tenant_id,
    storage_backend_from_env,
)

def default_experiment_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "experiments" / "experiments.json"


@dataclass
class VariantMetrics:
    successes: int = 0
    failures: int = 0

    @property
    def total(self) -> int:
        return self.successes + self.failures

    @property
    def success_rate(self) -> float:
        if self.total == 0:
            return 0.0
        return self.successes / self.total

    def record(self, success: bool) -> None:
        if success:
            self.successes += 1
        else:
            self.failures += 1

    def to_dict(self) -> dict[str, int]:
        return {"successes": self.successes, "failures": self.failures}

    @classmethod
    def from_dict(cls, payload: dict[str, object]) -> "VariantMetrics":
        return cls(
            successes=int(payload.get("successes", 0)),
            failures=int(payload.get("failures", 0)),
        )


@dataclass
class ExperimentState:
    baseline_variant: str
    active_variant: str
    variants: dict[str, VariantMetrics] = field(default_factory=dict)
    previous_variant: str | None = None

    def ensure_variant(self, variant_id: str) -> VariantMetrics:
        if variant_id not in self.variants:
            self.variants[variant_id] = VariantMetrics()
        return self.variants[variant_id]

    def to_dict(self) -> dict[str, object]:
        return {
            "baseline_variant": self.baseline_variant,
            "active_variant": self.active_variant,
            "previous_variant": self.previous_variant,
            "variants": {
                variant_id: metrics.to_dict()
                for variant_id, metrics in self.variants.items()
            },
        }

    @classmethod
    def from_dict(cls, payload: dict[str, object]) -> "ExperimentState":
        baseline = str(payload.get("baseline_variant") or "")
        active = str(payload.get("active_variant") or baseline)
        previous_variant = payload.get("previous_variant")
        raw_variants = payload.get("variants", {})
        variants: dict[str, VariantMetrics] = {}
        if isinstance(raw_variants, dict):
            for variant_id, metrics_payload in raw_variants.items():
                if isinstance(metrics_payload, dict):
                    variants[variant_id] = VariantMetrics.from_dict(metrics_payload)
        if baseline and baseline not in variants:
            variants[baseline] = VariantMetrics()
        if active and active not in variants:
            variants[active] = VariantMetrics()
        return cls(
            baseline_variant=baseline or active,
            active_variant=active or baseline,
            previous_variant=previous_variant if isinstance(previous_variant, str) else None,
            variants=variants,
        )


class ExperimentStore:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)

    def load(self) -> dict[str, ExperimentState]:
        if not self.path.exists():
            return {}
        raw = json.loads(self.path.read_text(encoding="utf-8"))
        experiments = raw.get("experiments", raw)
        if not isinstance(experiments, dict):
            return {}
        parsed: dict[str, ExperimentState] = {}
        for experiment_id, payload in experiments.items():
            if isinstance(payload, dict):
                parsed[experiment_id] = ExperimentState.from_dict(payload)
        return parsed

    def save(self, experiments: dict[str, ExperimentState]) -> None:
        payload = {
            "experiments": {
                experiment_id: state.to_dict()
                for experiment_id, state in experiments.items()
            }
        }
        atomic_write_text(
            self.path,
            json.dumps(payload, ensure_ascii=True, indent=2),
            encoding="utf-8",
        )

    def get(self, experiment_id: str) -> ExperimentState | None:
        return self.load().get(experiment_id)

    def upsert(self, experiment_id: str, state: ExperimentState) -> None:
        experiments = self.load()
        experiments[experiment_id] = state
        self.save(experiments)


class ExperimentStorePostgres:
    def __init__(self, dsn: str, *, tenant_id: str, schema: str = "skillos") -> None:
        self._dsn = dsn
        self._tenant_id = tenant_id
        self._schema = schema
        self._ensure_table()

    def load(self) -> dict[str, ExperimentState]:
        query = (
            f"SELECT experiment_id, baseline_variant, active_variant, "
            f"previous_variant, variants FROM {self._schema}.experiments "
            "WHERE tenant_id = %s"
        )
        with pg_connect(self._dsn) as conn:
            rows = conn.execute(query, (self._tenant_id,)).fetchall()
        parsed: dict[str, ExperimentState] = {}
        for row in rows:
            payload = {
                "baseline_variant": row.get("baseline_variant"),
                "active_variant": row.get("active_variant"),
                "previous_variant": row.get("previous_variant"),
                "variants": row.get("variants") or {},
            }
            parsed[str(row.get("experiment_id"))] = ExperimentState.from_dict(payload)
        return parsed

    def save(self, experiments: dict[str, ExperimentState]) -> None:
        with pg_connect(self._dsn) as conn:
            if not experiments:
                conn.execute(
                    f"DELETE FROM {self._schema}.experiments WHERE tenant_id = %s",
                    (self._tenant_id,),
                )
                return
            experiment_ids = list(experiments.keys())
            conn.execute(
                f"""
                DELETE FROM {self._schema}.experiments
                WHERE tenant_id = %s AND NOT (experiment_id = ANY(%s))
                """,
                (self._tenant_id, experiment_ids),
            )
            for experiment_id, state in experiments.items():
                conn.execute(
                    f"""
                    INSERT INTO {self._schema}.experiments (
                        tenant_id, experiment_id, baseline_variant, active_variant,
                        previous_variant, variants
                    ) VALUES (%s, %s, %s, %s, %s, %s::jsonb)
                    ON CONFLICT (tenant_id, experiment_id) DO UPDATE SET
                        baseline_variant = EXCLUDED.baseline_variant,
                        active_variant = EXCLUDED.active_variant,
                        previous_variant = EXCLUDED.previous_variant,
                        variants = EXCLUDED.variants,
                        updated_at = now()
                    """,
                    (
                        self._tenant_id,
                        experiment_id,
                        state.baseline_variant,
                        state.active_variant,
                        state.previous_variant,
                        json.dumps(state.to_dict().get("variants", {}), ensure_ascii=True),
                    ),
                )

    def get(self, experiment_id: str) -> ExperimentState | None:
        query = (
            f"SELECT baseline_variant, active_variant, previous_variant, variants "
            f"FROM {self._schema}.experiments WHERE tenant_id = %s AND experiment_id = %s"
        )
        with pg_connect(self._dsn) as conn:
            row = conn.execute(query, (self._tenant_id, experiment_id)).fetchone()
        if not row:
            return None
        payload = {
            "baseline_variant": row.get("baseline_variant"),
            "active_variant": row.get("active_variant"),
            "previous_variant": row.get("previous_variant"),
            "variants": row.get("variants") or {},
        }
        return ExperimentState.from_dict(payload)

    def upsert(self, experiment_id: str, state: ExperimentState) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(
                f"""
                INSERT INTO {self._schema}.experiments (
                    tenant_id, experiment_id, baseline_variant, active_variant,
                    previous_variant, variants
                ) VALUES (%s, %s, %s, %s, %s, %s::jsonb)
                ON CONFLICT (tenant_id, experiment_id) DO UPDATE SET
                    baseline_variant = EXCLUDED.baseline_variant,
                    active_variant = EXCLUDED.active_variant,
                    previous_variant = EXCLUDED.previous_variant,
                    variants = EXCLUDED.variants,
                    updated_at = now()
                """,
                (
                    self._tenant_id,
                    experiment_id,
                    state.baseline_variant,
                    state.active_variant,
                    state.previous_variant,
                    json.dumps(state.to_dict().get("variants", {}), ensure_ascii=True),
                ),
            )

    def _ensure_table(self) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {self._schema}")
            conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self._schema}.experiments (
                    tenant_id TEXT NOT NULL,
                    experiment_id TEXT NOT NULL,
                    baseline_variant TEXT NOT NULL,
                    active_variant TEXT NOT NULL,
                    previous_variant TEXT,
                    variants JSONB NOT NULL,
                    updated_at TIMESTAMPTZ NOT NULL DEFAULT now(),
                    PRIMARY KEY (tenant_id, experiment_id)
                )
                """
            )


def experiment_store_from_env(root: Path) -> ExperimentStore | ExperimentStorePostgres:
    config = storage_backend_from_env()
    if config.backend != "postgres":
        return ExperimentStore(default_experiment_path(root))
    dsn = require_postgres_dsn(config, context="experiments")
    tenant_id = resolve_tenant_id(Path(root))
    return ExperimentStorePostgres(
        dsn,
        tenant_id=tenant_id,
        schema=config.postgres_schema,
    )


class ExperimentTracker:
    def __init__(
        self,
        store: ExperimentStore,
        experiment_id: str,
        variants: list[str],
        *,
        baseline_variant: str | None = None,
    ) -> None:
        if not variants:
            raise ValueError("At least one variant is required")
        self._store = store
        self._experiment_id = experiment_id
        self._variants = list(dict.fromkeys(variants))
        if not self._variants:
            raise ValueError("At least one variant is required")
        self._baseline_variant = baseline_variant or self._variants[0]
        if self._baseline_variant not in self._variants:
            raise ValueError("Baseline variant must be included in variants")
        self._ensure_state()

    @property
    def experiment_id(self) -> str:
        return self._experiment_id

    def assign_variant(self, user_id: str) -> str:
        digest = hashlib.sha256(
            f"{self._experiment_id}:{user_id}".encode("utf-8")
        ).hexdigest()
        index = int(digest, 16) % len(self._variants)
        return self._variants[index]

    def record_outcome(self, variant_id: str, success: bool) -> VariantMetrics:
        state = self._load_state()
        metrics = state.ensure_variant(variant_id)
        metrics.record(success)
        self._store.upsert(self._experiment_id, state)
        return metrics

    def metrics_for(self, variant_id: str) -> VariantMetrics:
        state = self._load_state()
        return state.ensure_variant(variant_id)

    def _ensure_state(self) -> None:
        state = self._store.get(self._experiment_id)
        if not state:
            state = ExperimentState(
                baseline_variant=self._baseline_variant,
                active_variant=self._baseline_variant,
                variants={variant: VariantMetrics() for variant in self._variants},
            )
            self._store.upsert(self._experiment_id, state)
            return
        for variant_id in self._variants:
            state.ensure_variant(variant_id)
        if not state.baseline_variant:
            state.baseline_variant = self._baseline_variant
        if not state.active_variant:
            state.active_variant = state.baseline_variant
        self._store.upsert(self._experiment_id, state)

    def _load_state(self) -> ExperimentState:
        state = self._store.get(self._experiment_id)
        if not state:
            state = ExperimentState(
                baseline_variant=self._baseline_variant,
                active_variant=self._baseline_variant,
                variants={variant: VariantMetrics() for variant in self._variants},
            )
        for variant_id in self._variants:
            state.ensure_variant(variant_id)
        return state



from pathlib import Path

import yaml
from pydantic import ValidationError

from skillos.skills.errors import SkillValidationError
from skillos.skills.models import SkillMetadata

_SKILL_CACHE: dict[Path, tuple[int, SkillMetadata]] = {}


def load_skill_file(path: Path) -> SkillMetadata:
    path = Path(path)
    try:
        mtime = path.stat().st_mtime_ns
    except FileNotFoundError:
        _SKILL_CACHE.pop(path, None)
        raise

    cached = _SKILL_CACHE.get(path)
    if cached and cached[0] == mtime:
        return cached[1]
    try:
        raw = yaml.safe_load(path.read_text(encoding="utf-8"))
    except FileNotFoundError:
        raise
    except Exception as exc:  # pragma: no cover - defensive
        raise SkillValidationError(f"Unable to read skill YAML: {exc}") from exc

    if not isinstance(raw, dict):
        raise SkillValidationError("Skill YAML must be a mapping")

    try:
        metadata = SkillMetadata.model_validate(raw)
    except ValidationError as exc:
        raise SkillValidationError(_format_validation_errors(exc)) from exc
    _SKILL_CACHE[path] = (mtime, metadata)
    return metadata


def _format_validation_errors(error: ValidationError) -> str:
    parts: list[str] = []
    for item in error.errors():
        loc = ".".join(str(piece) for piece in item.get("loc", [])) or "skill"
        parts.append(f"{loc}: {item.get('msg')}")
    return "; ".join(parts)



from __future__ import annotations

from dataclasses import dataclass
import hashlib
import json
from typing import Iterable

from skillos.cache import CacheBackend, CacheConfig, cache_backend_from_env, cache_config_from_env
from skillos.routing import RoutingResult, SkillCandidate
from skillos.tenancy import tenant_id_from_path


@dataclass(frozen=True)
class RoutingCacheConfig:
    enabled: bool
    ttl_seconds: int
    prefix: str


class RoutingCache:
    def __init__(
        self,
        backend: CacheBackend,
        *,
        tenant_id: str,
        ttl_seconds: int,
        prefix: str,
    ) -> None:
        self._backend = backend
        self._tenant_id = tenant_id
        self._ttl_seconds = ttl_seconds
        self._prefix = prefix

    def get(self, query: str, tags: Iterable[str] | None = None) -> RoutingResult | None:
        key = self._key(query, tags)
        raw = self._backend.get(key)
        if not raw:
            return None
        try:
            payload = json.loads(raw)
        except json.JSONDecodeError:
            return None
        return _routing_result_from_dict(payload)

    def set(
        self,
        query: str,
        tags: Iterable[str] | None,
        result: RoutingResult,
    ) -> None:
        key = self._key(query, tags)
        payload = json.dumps(
            _routing_result_to_dict(result),
            ensure_ascii=True,
        )
        self._backend.set(key, payload, self._ttl_seconds)

    def _key(self, query: str, tags: Iterable[str] | None) -> str:
        query_hash = _hash_text(query)
        tag_hash = _hash_tags(tags)
        return f"{self._prefix}:tenant:{self._tenant_id}:routing:{query_hash}:{tag_hash}"


def routing_cache_from_env(root_path) -> RoutingCache | None:
    config = cache_config_from_env()
    if not config.enabled:
        return None
    backend = cache_backend_from_env(config)
    if backend is None:
        return None
    tenant_id = tenant_id_from_path(root_path) or "default"
    routing_config = RoutingCacheConfig(
        enabled=config.enabled,
        ttl_seconds=config.ttl_seconds,
        prefix=config.prefix,
    )
    return RoutingCache(
        backend,
        tenant_id=tenant_id,
        ttl_seconds=routing_config.ttl_seconds,
        prefix=routing_config.prefix,
    )


def _routing_result_to_dict(result: RoutingResult) -> dict[str, object]:
    return {
        "status": result.status,
        "skill_id": result.skill_id,
        "internal_skill_id": result.internal_skill_id,
        "confidence": result.confidence,
        "alternatives": result.alternatives,
        "candidates": [
            {
                "skill_id": candidate.skill_id,
                "score": candidate.score,
                "keyword_score": candidate.keyword_score,
                "semantic_score": candidate.semantic_score,
                "internal_id": candidate.internal_id,
            }
            for candidate in result.candidates
        ],
    }


def _routing_result_from_dict(payload: dict[str, object]) -> RoutingResult | None:
    if not isinstance(payload, dict):
        return None
    candidates_raw = payload.get("candidates", [])
    candidates: list[SkillCandidate] = []
    if isinstance(candidates_raw, list):
        for item in candidates_raw:
            if not isinstance(item, dict):
                continue
            candidates.append(
                SkillCandidate(
                    skill_id=str(item.get("skill_id")),
                    score=float(item.get("score", 0.0)),
                    keyword_score=int(item.get("keyword_score", 0)),
                    semantic_score=float(item.get("semantic_score", 0.0)),
                    internal_id=item.get("internal_id"),
                )
            )
    return RoutingResult(
        status=str(payload.get("status")),
        skill_id=payload.get("skill_id"),
        internal_skill_id=payload.get("internal_skill_id"),
        confidence=float(payload.get("confidence", 0.0)),
        candidates=candidates,
        alternatives=[
            str(item) for item in payload.get("alternatives", []) if item is not None
        ],
    )


def _hash_text(value: str) -> str:
    return hashlib.sha256(value.encode("utf-8")).hexdigest()


def _hash_tags(tags: Iterable[str] | None) -> str:
    if not tags:
        return "none"
    normalized = sorted(tag.strip().lower() for tag in tags if tag.strip())
    if not normalized:
        return "none"
    return _hash_text(",".join(normalized))



from __future__ import annotations

import re


_PIPELINE_SPLIT_RE = re.compile(r"\bthen\b|\bafter\b|->", re.IGNORECASE)
_PARALLEL_SPLIT_RE = re.compile(r"\band\b|&", re.IGNORECASE)


def select_mode(query: str) -> str:
    if _PIPELINE_SPLIT_RE.search(query):
        return "pipeline"
    if _PARALLEL_SPLIT_RE.search(query):
        return "parallel"
    return "single"


def split_query(query: str, mode: str) -> list[str]:
    normalized = (mode or "single").strip().lower()
    if normalized == "auto":
        normalized = select_mode(query)
    if normalized == "pipeline":
        parts = _PIPELINE_SPLIT_RE.split(query)
    elif normalized == "parallel":
        parts = _PARALLEL_SPLIT_RE.split(query)
    else:
        return [query.strip()] if query.strip() else []
    cleaned = [part.strip() for part in parts if part.strip()]
    return cleaned if len(cleaned) > 1 else [query.strip()] if query.strip() else []



from __future__ import annotations
import json
from pathlib import Path
import click

from skillos.marketplace import (
    default_catalog_path,
    load_marketplace_catalog,
    get_package,
    install_package,
    uninstall_package,
    MarketplaceError,
    CatalogError,
)
from skillos.skills.paths import default_skills_root
from skillos.telemetry import EventLogger, default_log_path, new_request_id

@click.group("marketplace")
def marketplace() -> None:
    """Browse and install community skills."""

@marketplace.command("browse")
@click.option(
    "--catalog",
    "catalog_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=default_catalog_path(),
    show_default=True,
)
@click.option("--query", default=None)
@click.option("--tag", default=None)
def marketplace_browse(
    catalog_path: Path, query: str | None, tag: str | None
) -> None:
    """List packages from the marketplace catalog."""
    from skillos.marketplace import filter_packages  # keeping loose import if not used elsewhere, or move up
    
    try:
        packages = load_marketplace_catalog(catalog_path)
    except (CatalogError, FileNotFoundError) as exc:
        raise click.ClickException(str(exc)) from exc

    filtered = filter_packages(packages, query=query, tag=tag)
    if not filtered:
        click.echo("no_packages_found")
        return

    for package in filtered:
        click.echo(f"{package.id} {package.version} {package.name}")

@marketplace.command("show")
@click.argument("package_id")
@click.option(
    "--catalog",
    "catalog_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=default_catalog_path(),
    show_default=True,
)
def marketplace_show(package_id: str, catalog_path: Path) -> None:
    """Show details for a marketplace package."""
    try:
        packages = load_marketplace_catalog(catalog_path)
        package = get_package(packages, package_id)
    except (MarketplaceError, FileNotFoundError) as exc:
        raise click.ClickException(str(exc)) from exc

    click.echo(json.dumps(package.model_dump(), indent=2, sort_keys=True))

@marketplace.command("install")
@click.argument("package_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option(
    "--catalog",
    "catalog_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=default_catalog_path(),
    show_default=True,
)
def marketplace_install(
    package_id: str, root_path: Path, catalog_path: Path
) -> None:
    """Install a marketplace package and dependencies."""
    logger = EventLogger(default_log_path(root_path), request_id=new_request_id())
    try:
        installed = install_package(
            package_id,
            root_path,
            catalog_path,
            logger=logger,
        )
    except MarketplaceError as exc:
        raise click.ClickException(str(exc)) from exc

    for pkg_id in installed:
        click.echo(f"installed: {pkg_id}")

@marketplace.command("uninstall")
@click.argument("package_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option("--prune/--no-prune", default=True, show_default=True)
def marketplace_uninstall(
    package_id: str, root_path: Path, prune: bool
) -> None:
    """Uninstall a marketplace package."""
    try:
        removed = uninstall_package(package_id, root_path, prune_orphans=prune)
    except MarketplaceError as exc:
        raise click.ClickException(str(exc)) from exc

    for pkg_id in removed:
        click.echo(f"removed: {pkg_id}")



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Protocol

from skillos.sdk import Context


class NodeProtocol(Protocol):
    def __call__(self, state: dict, ctx: Context) -> dict: ...


@dataclass
class FlowState:
    data: dict = field(default_factory=dict)
    history: list[str] = field(default_factory=list)
    next_node: str | None = None
    terminated: bool = False


class SkillFlow:
    """
    A lightweight state machine for defining cyclic or multi-step skills.
    Inspired by LangGraph but adapted for SkillOS's functional paradigm.
    """

    def __init__(self, name: str, entry_point: str = "start"):
        self.name = name
        self.nodes: Dict[str, NodeProtocol] = {}
        self.edges: Dict[str, Callable[[dict], str]] = {}
        self.entry_point = entry_point
        self._on_start = None

    def node(self, name: str) -> Callable:
        """Decorator to register a node function."""
        def decorator(func: NodeProtocol) -> NodeProtocol:
            self.nodes[name] = func
            return func
        return decorator

    def add_edge(self, from_node: str, condition: Callable[[dict], str]):
        """Define transition logic from a node."""
        self.edges[from_node] = condition

    def start(self, func: NodeProtocol):
        """Decorator for the entry point node."""
        self.nodes[self.entry_point] = func
        return func

    def run(self, initial_payload: Any, ctx: Context = None) -> Any:
        # Initialize state
        state = initial_payload if isinstance(initial_payload, dict) else {"payload": initial_payload}
        current_node = self.entry_point
        steps = 0
        max_steps = 50  # Safety limit

        if ctx:
            ctx.log(f"Flow {self.name} started at {current_node}")

        while current_node and steps < max_steps:
            if current_node not in self.nodes:
                raise ValueError(f"Flow terminated: Node '{current_node}' not found.")

            node_func = self.nodes[current_node]
            
            # Execute node
            # We support functions that take (state, ctx) or just (state)
            sig = inspect.signature(node_func)
            if "ctx" in sig.parameters:
                new_state = node_func(state, ctx=ctx)
            else:
                new_state = node_func(state)
            
            # Update state (merge or replace? LangGraph typically passes state. simple replace for now)
            if new_state is not None:
                if isinstance(new_state, dict):
                     state.update(new_state)
                else:
                     state["result"] = new_state

            steps += 1
            
            # Determine next node
            if current_node in self.edges:
                transition_logic = self.edges[current_node]
                next_node = transition_logic(state)
            else:
                next_node = None # End of flow

            if ctx:
                ctx.log(f"Flow step {steps}: {current_node} -> {next_node}")
            
            current_node = next_node
            
            if next_node == "__end__":
                break
        
        if steps >= max_steps:
             raise RecursionError(f"Flow '{self.name}' exceeded maximum steps ({max_steps})")

        return state



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
import json
from pathlib import Path

from skillos.storage import atomic_write_text
from skillos.tenancy import resolve_tenant_root
from skillos.storage_backend import (
    pg_connect,
    require_postgres_dsn,
    resolve_tenant_id,
    storage_backend_from_env,
)

def default_preferences_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "suggestions" / "preferences.json"


def default_suggestions_path(root: Path) -> Path:
    root_path = resolve_tenant_root(root)
    return root_path / "suggestions" / "suggestions.json"


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _ensure_utc(value: datetime) -> datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=timezone.utc)
    return value.astimezone(timezone.utc)


def _parse_datetime(value: str | None) -> datetime | None:
    if not value:
        return None
    normalized = value
    if normalized.endswith("Z"):
        normalized = normalized[:-1] + "+00:00"
    parsed = datetime.fromisoformat(normalized)
    return _ensure_utc(parsed)


def _format_datetime(value: datetime | None) -> str | None:
    if value is None:
        return None
    return _ensure_utc(value).isoformat()


@dataclass
class SuggestionPreferences:
    opt_in: bool = False
    max_per_day: int = 3
    min_interval_minutes: int = 60
    cooldown_minutes_on_dismiss: int = 120
    snoozed_until: datetime | None = None

    def to_dict(self) -> dict[str, object]:
        return {
            "opt_in": self.opt_in,
            "max_per_day": self.max_per_day,
            "min_interval_minutes": self.min_interval_minutes,
            "cooldown_minutes_on_dismiss": self.cooldown_minutes_on_dismiss,
            "snoozed_until": _format_datetime(self.snoozed_until),
        }

    @classmethod
    def from_dict(cls, payload: dict[str, object]) -> "SuggestionPreferences":
        opt_in = bool(payload.get("opt_in", False))
        max_per_day = int(payload.get("max_per_day", 3))
        min_interval_minutes = int(payload.get("min_interval_minutes", 60))
        cooldown_minutes = int(payload.get("cooldown_minutes_on_dismiss", 120))
        snoozed_until_raw = payload.get("snoozed_until")
        snoozed_until = (
            _parse_datetime(str(snoozed_until_raw)) if snoozed_until_raw else None
        )
        return cls(
            opt_in=opt_in,
            max_per_day=max_per_day,
            min_interval_minutes=min_interval_minutes,
            cooldown_minutes_on_dismiss=cooldown_minutes,
            snoozed_until=snoozed_until,
        )


@dataclass
class SuggestionRecord:
    suggestion_id: str
    source: str
    summary: str
    message: str
    created_at: datetime
    status: str = "created"
    dismissed_at: datetime | None = None

    def to_dict(self) -> dict[str, object]:
        return {
            "suggestion_id": self.suggestion_id,
            "source": self.source,
            "summary": self.summary,
            "message": self.message,
            "created_at": _format_datetime(self.created_at),
            "status": self.status,
            "dismissed_at": _format_datetime(self.dismissed_at),
        }

    @classmethod
    def from_dict(cls, payload: dict[str, object]) -> "SuggestionRecord":
        raw_created_at = payload.get("created_at")
        created_at = (
            _parse_datetime(str(raw_created_at)) if raw_created_at else None
        )
        return cls(
            suggestion_id=str(payload.get("suggestion_id", "")),
            source=str(payload.get("source", "")),
            summary=str(payload.get("summary", "")),
            message=str(payload.get("message", "")),
            created_at=created_at or _utc_now(),
            status=str(payload.get("status", "created")),
            dismissed_at=_parse_datetime(str(payload.get("dismissed_at")))
            if payload.get("dismissed_at")
            else None,
        )


class SuggestionPreferencesStore:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)

    def load(self) -> SuggestionPreferences:
        if not self.path.exists():
            return SuggestionPreferences()
        raw = json.loads(self.path.read_text(encoding="utf-8"))
        if not isinstance(raw, dict):
            return SuggestionPreferences()
        return SuggestionPreferences.from_dict(raw)

    def save(self, preferences: SuggestionPreferences) -> None:
        atomic_write_text(
            self.path,
            json.dumps(preferences.to_dict(), ensure_ascii=True, indent=2),
            encoding="utf-8",
        )


class SuggestionStore:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)

    def load(self) -> list[SuggestionRecord]:
        if not self.path.exists():
            return []
        raw = json.loads(self.path.read_text(encoding="utf-8"))
        if isinstance(raw, dict):
            raw = raw.get("suggestions", [])
        if not isinstance(raw, list):
            return []
        records: list[SuggestionRecord] = []
        for item in raw:
            if isinstance(item, dict):
                records.append(SuggestionRecord.from_dict(item))
        return records

    def save(self, records: list[SuggestionRecord]) -> None:
        payload = {"suggestions": [record.to_dict() for record in records]}
        atomic_write_text(
            self.path,
            json.dumps(payload, ensure_ascii=True, indent=2),
            encoding="utf-8",
        )


class SuggestionPreferencesStorePostgres:
    def __init__(self, dsn: str, *, tenant_id: str, schema: str = "skillos") -> None:
        self._dsn = dsn
        self._tenant_id = tenant_id
        self._schema = schema
        self._ensure_table()

    def load(self) -> SuggestionPreferences:
        query = (
            f"SELECT opt_in, max_per_day, min_interval_minutes, "
            f"cooldown_minutes_on_dismiss, snoozed_until "
            f"FROM {self._schema}.suggestion_preferences WHERE tenant_id = %s"
        )
        with pg_connect(self._dsn) as conn:
            row = conn.execute(query, (self._tenant_id,)).fetchone()
        if not row:
            return SuggestionPreferences()
        return SuggestionPreferences(
            opt_in=bool(row.get("opt_in", False)),
            max_per_day=int(row.get("max_per_day", 3)),
            min_interval_minutes=int(row.get("min_interval_minutes", 60)),
            cooldown_minutes_on_dismiss=int(row.get("cooldown_minutes_on_dismiss", 120)),
            snoozed_until=_ensure_utc(row["snoozed_until"])
            if row.get("snoozed_until")
            else None,
        )

    def save(self, preferences: SuggestionPreferences) -> None:
        query = (
            f"""
            INSERT INTO {self._schema}.suggestion_preferences (
                tenant_id, opt_in, max_per_day, min_interval_minutes,
                cooldown_minutes_on_dismiss, snoozed_until
            ) VALUES (%s, %s, %s, %s, %s, %s)
            ON CONFLICT (tenant_id) DO UPDATE SET
                opt_in = EXCLUDED.opt_in,
                max_per_day = EXCLUDED.max_per_day,
                min_interval_minutes = EXCLUDED.min_interval_minutes,
                cooldown_minutes_on_dismiss = EXCLUDED.cooldown_minutes_on_dismiss,
                snoozed_until = EXCLUDED.snoozed_until,
                updated_at = now()
            """
        )
        with pg_connect(self._dsn) as conn:
            conn.execute(
                query,
                (
                    self._tenant_id,
                    preferences.opt_in,
                    preferences.max_per_day,
                    preferences.min_interval_minutes,
                    preferences.cooldown_minutes_on_dismiss,
                    _ensure_utc(preferences.snoozed_until)
                    if preferences.snoozed_until
                    else None,
                ),
            )

    def _ensure_table(self) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {self._schema}")
            conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self._schema}.suggestion_preferences (
                    tenant_id TEXT PRIMARY KEY,
                    opt_in BOOLEAN NOT NULL,
                    max_per_day INTEGER NOT NULL,
                    min_interval_minutes INTEGER NOT NULL,
                    cooldown_minutes_on_dismiss INTEGER NOT NULL,
                    snoozed_until TIMESTAMPTZ,
                    updated_at TIMESTAMPTZ NOT NULL DEFAULT now()
                )
                """
            )


class SuggestionStorePostgres:
    def __init__(self, dsn: str, *, tenant_id: str, schema: str = "skillos") -> None:
        self._dsn = dsn
        self._tenant_id = tenant_id
        self._schema = schema
        self._ensure_table()

    def load(self) -> list[SuggestionRecord]:
        query = (
            f"SELECT * FROM {self._schema}.suggestions "
            "WHERE tenant_id = %s ORDER BY created_at"
        )
        with pg_connect(self._dsn) as conn:
            rows = conn.execute(query, (self._tenant_id,)).fetchall()
        records: list[SuggestionRecord] = []
        for row in rows:
            records.append(
                SuggestionRecord(
                    suggestion_id=str(row.get("suggestion_id") or ""),
                    source=str(row.get("source") or ""),
                    summary=str(row.get("summary") or ""),
                    message=str(row.get("message") or ""),
                    created_at=_ensure_utc(row["created_at"])
                    if row.get("created_at")
                    else _utc_now(),
                    status=str(row.get("status") or "created"),
                    dismissed_at=_ensure_utc(row["dismissed_at"])
                    if row.get("dismissed_at")
                    else None,
                )
            )
        return records

    def save(self, records: list[SuggestionRecord]) -> None:
        with pg_connect(self._dsn) as conn:
            if not records:
                conn.execute(
                    f"DELETE FROM {self._schema}.suggestions WHERE tenant_id = %s",
                    (self._tenant_id,),
                )
                return
            suggestion_ids = [record.suggestion_id for record in records]
            conn.execute(
                f"""
                DELETE FROM {self._schema}.suggestions
                WHERE tenant_id = %s AND NOT (suggestion_id = ANY(%s))
                """,
                (self._tenant_id, suggestion_ids),
            )
            for record in records:
                conn.execute(
                    f"""
                    INSERT INTO {self._schema}.suggestions (
                        tenant_id, suggestion_id, source, summary, message,
                        created_at, status, dismissed_at
                    ) VALUES (%s, %s, %s, %s, %s, %s, %s, %s)
                    ON CONFLICT (tenant_id, suggestion_id) DO UPDATE SET
                        source = EXCLUDED.source,
                        summary = EXCLUDED.summary,
                        message = EXCLUDED.message,
                        created_at = EXCLUDED.created_at,
                        status = EXCLUDED.status,
                        dismissed_at = EXCLUDED.dismissed_at
                    """,
                    (
                        self._tenant_id,
                        record.suggestion_id,
                        record.source,
                        record.summary,
                        record.message,
                        _ensure_utc(record.created_at),
                        record.status,
                        _ensure_utc(record.dismissed_at)
                        if record.dismissed_at
                        else None,
                    ),
                )

    def _ensure_table(self) -> None:
        with pg_connect(self._dsn) as conn:
            conn.execute(f"CREATE SCHEMA IF NOT EXISTS {self._schema}")
            conn.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self._schema}.suggestions (
                    tenant_id TEXT NOT NULL,
                    suggestion_id TEXT NOT NULL,
                    source TEXT NOT NULL,
                    summary TEXT NOT NULL,
                    message TEXT NOT NULL,
                    created_at TIMESTAMPTZ NOT NULL,
                    status TEXT NOT NULL,
                    dismissed_at TIMESTAMPTZ,
                    PRIMARY KEY (tenant_id, suggestion_id)
                )
                """
            )


def preferences_store_from_env(
    root: Path,
) -> SuggestionPreferencesStore | SuggestionPreferencesStorePostgres:
    config = storage_backend_from_env()
    if config.backend != "postgres":
        return SuggestionPreferencesStore(default_preferences_path(root))
    dsn = require_postgres_dsn(config, context="suggestions")
    tenant_id = resolve_tenant_id(Path(root))
    return SuggestionPreferencesStorePostgres(
        dsn,
        tenant_id=tenant_id,
        schema=config.postgres_schema,
    )


def suggestion_store_from_env(
    root: Path,
) -> SuggestionStore | SuggestionStorePostgres:
    config = storage_backend_from_env()
    if config.backend != "postgres":
        return SuggestionStore(default_suggestions_path(root))
    dsn = require_postgres_dsn(config, context="suggestions")
    tenant_id = resolve_tenant_id(Path(root))
    return SuggestionStorePostgres(
        dsn,
        tenant_id=tenant_id,
        schema=config.postgres_schema,
    )



from __future__ import annotations

from dataclasses import dataclass

from skillos.execution_planner import ExecutionPlan


@dataclass(frozen=True)
class ExecutionPreview:
    plan_id: str
    affected_entities: list[str]

    def to_dict(self) -> dict[str, object]:
        return {
            "plan_id": self.plan_id,
            "affected_entities": self.affected_entities,
        }


def render_preview(plan: ExecutionPlan) -> ExecutionPreview:
    affected_entities = [plan.skill_id]
    return ExecutionPreview(plan_id=plan.plan_id, affected_entities=affected_entities)



from __future__ import annotations
from typing import Dict, Type
from skillos.kernel.base import ExecutionKernel
from skillos.kernel.local import LocalExecutionKernel
from skillos.skills.models import SkillMetadata

_KERNELS: Dict[str, ExecutionKernel] = {}

def get_kernel(metadata: SkillMetadata, root_path: str | None = None) -> ExecutionKernel:
    """Factory to get the appropriate kernel for a skill."""
    mode = getattr(metadata, "execution_mode", "local")
    
    if mode not in _KERNELS:
        if mode == "local":
            _KERNELS[mode] = LocalExecutionKernel(root_path)
        elif mode == "docker":
            raise NotImplementedError("Docker execution kernel is not yet implemented.")
        elif mode == "remote":
            raise NotImplementedError("Remote execution kernel is not yet implemented.")
        else:
            raise ValueError(f"Unknown execution mode: {mode}")
            
    return _KERNELS[mode]



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
import json
import re
from pathlib import Path
from typing import Callable
from concurrent.futures import ThreadPoolExecutor, TimeoutError as FutureTimeoutError

import httpx

from skillos.skills.models import SkillEvalConfig, SkillMetadata
from skillos.skills.registry import SkillRegistry
from skillos.storage import atomic_write_text
from skillos.testing import mock_external_apis

_NUMBER_RE = re.compile(r"[-+]?(?:\d+\.?\d*|\d*\.?\d+)")


class SkillEvalError(ValueError):
    pass


@dataclass(frozen=True)
class EvalCaseResult:
    input: str
    expected: str | None
    match: str
    passed: bool
    output: str
    details: str | None = None


@dataclass(frozen=True)
class EvalRunResult:
    skill_id: str
    total: int
    passed: int
    success_rate: float
    pass_threshold: float
    cases: list[EvalCaseResult]

    @property
    def ok(self) -> bool:
        return self.success_rate >= self.pass_threshold

    def to_dict(self) -> dict[str, object]:
        return {
            "skill_id": self.skill_id,
            "total": self.total,
            "passed": self.passed,
            "success_rate": self.success_rate,
            "pass_threshold": self.pass_threshold,
            "ok": self.ok,
            "cases": [
                {
                    "input": case.input,
                    "expected": case.expected,
                    "match": case.match,
                    "passed": case.passed,
                    "output": case.output,
                    "details": case.details,
                }
                for case in self.cases
            ],
        }


def run_skill_eval(
    skill_id: str,
    root: Path,
    *,
    eval_config: SkillEvalConfig | None = None,
    mock_handler: Callable[[httpx.Request], httpx.Response] | None = None,
) -> EvalRunResult:
    registry = SkillRegistry(root)
    registry.load_all()
    metadata = registry.get(skill_id)
    if not metadata:
        raise KeyError(f"Unknown skill: {skill_id}")

    config = eval_config or metadata.eval
    if not config or not config.cases:
        raise SkillEvalError("eval_not_configured")

    results: list[EvalCaseResult] = []
    cases = config.cases
    if config.max_cases is not None:
        cases = cases[: config.max_cases]
    timeout_seconds = config.timeout_seconds
    with mock_external_apis(handler=mock_handler):
        for case in cases:
            output_text, details = _execute_case(
                registry,
                skill_id,
                case.input,
                timeout_seconds,
            )
            if output_text is None:
                results.append(
                    EvalCaseResult(
                        input=case.input,
                        expected=case.expected,
                        match=case.match,
                        passed=False,
                        output="",
                        details=details or "execution_failed",
                    )
                )
                if config.fail_fast:
                    break
                continue
            passed, details = _match_case(case, output_text)
            results.append(
                EvalCaseResult(
                    input=case.input,
                    expected=case.expected,
                    match=case.match,
                    passed=passed,
                    output=output_text,
                    details=details,
                )
            )
            if config.fail_fast and not passed:
                break

    passed_count = sum(1 for item in results if item.passed)
    total = len(results)
    success_rate = passed_count / total if total else 0.0
    return EvalRunResult(
        skill_id=skill_id,
        total=total,
        passed=passed_count,
        success_rate=round(success_rate, 3),
        pass_threshold=config.pass_threshold,
        cases=results,
    )


def _execute_case(
    registry: SkillRegistry,
    skill_id: str,
    payload: str,
    timeout_seconds: float | None,
) -> tuple[str | None, str | None]:
    if timeout_seconds is None:
        return str(registry.execute(skill_id, payload=payload)), None

    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(registry.execute, skill_id, payload=payload)
        try:
            return str(future.result(timeout=timeout_seconds)), None
        except FutureTimeoutError:
            future.cancel()
            return None, "timeout"
        except Exception as exc:
            return None, exc.__class__.__name__


def default_eval_result_path(root: Path, skill_id: str) -> Path:
    safe_id = skill_id.replace("/", "_")
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")
    return Path(root) / "runtime" / "evals" / f"{safe_id}-{timestamp}.json"


def save_eval_result(
    result: EvalRunResult,
    root: Path,
    *,
    output_path: Path | None = None,
) -> Path:
    destination = output_path or default_eval_result_path(root, result.skill_id)
    destination.parent.mkdir(parents=True, exist_ok=True)
    atomic_write_text(
        destination,
        json.dumps(result.to_dict(), ensure_ascii=True, indent=2),
        encoding="utf-8",
    )
    return destination


def _match_case(case, output: str) -> tuple[bool, str | None]:
    match = case.match
    if match == "equals":
        expected = str(case.expected).strip() if case.expected is not None else ""
        ok = output.strip() == expected
        return ok, None if ok else f"expected '{expected}'"
    if match == "contains":
        expected = str(case.expected) if case.expected is not None else ""
        ok = expected in output
        return ok, None if ok else f"missing '{expected}'"
    if match == "regex":
        pattern = str(case.expected) if case.expected is not None else ""
        ok = re.search(pattern, output) is not None
        return ok, None if ok else f"regex '{pattern}' not matched"
    if match == "regex_numeric":
        pattern = str(case.expected) if case.expected is not None else ""
        found = re.search(pattern, output)
        if not found:
            return False, f"regex '{pattern}' not matched"
        group_index = case.group if case.group is not None else 1
        try:
            token = found.group(group_index)
        except IndexError:
            return False, f"group {group_index} not found"
        try:
            number = float(token)
        except ValueError:
            return False, f"non_numeric_group '{token}'"
        low, high = case.range
        ok = low <= number <= high
        return ok, None if ok else f"number {number} not in [{low}, {high}]"
    if match == "numeric_range":
        number = _extract_number(output)
        if number is None:
            return False, "no_number_found"
        low, high = case.range
        ok = low <= number <= high
        return ok, None if ok else f"number {number} not in [{low}, {high}]"
    return False, "unknown_matcher"


def _extract_number(text: str) -> float | None:
    match = _NUMBER_RE.search(text)
    if not match:
        return None
    try:
        return float(match.group(0))
    except ValueError:
        return None



from __future__ import annotations

from dataclasses import dataclass
import os
import time
from typing import Protocol


class CacheBackend(Protocol):
    def get(self, key: str) -> str | None:
        ...

    def set(self, key: str, value: str, ttl_seconds: int | None) -> None:
        ...


@dataclass(frozen=True)
class CacheConfig:
    enabled: bool = False
    ttl_seconds: int = 60
    prefix: str = "skillos"
    redis_url: str | None = None


class MemoryCache:
    def __init__(self) -> None:
        self._entries: dict[str, tuple[str, float | None]] = {}

    def get(self, key: str) -> str | None:
        entry = self._entries.get(key)
        if not entry:
            return None
        value, expires_at = entry
        if expires_at is not None and time.monotonic() >= expires_at:
            self._entries.pop(key, None)
            return None
        return value

    def set(self, key: str, value: str, ttl_seconds: int | None) -> None:
        expires_at = (
            time.monotonic() + ttl_seconds if ttl_seconds is not None else None
        )
        self._entries[key] = (value, expires_at)


class RedisCache:
    def __init__(self, url: str) -> None:
        import redis

        self._client = redis.Redis.from_url(url)

    def get(self, key: str) -> str | None:
        raw = self._client.get(key)
        if raw is None:
            return None
        if isinstance(raw, bytes):
            return raw.decode("utf-8")
        return str(raw)

    def set(self, key: str, value: str, ttl_seconds: int | None) -> None:
        if ttl_seconds is None:
            self._client.set(key, value)
            return
        self._client.setex(key, ttl_seconds, value)


def cache_config_from_env() -> CacheConfig:
    enabled = _env_bool("SKILLOS_CACHE_ENABLED", False)
    ttl_seconds = _env_int("SKILLOS_CACHE_TTL_SECONDS", 60)
    prefix = os.getenv("SKILLOS_CACHE_PREFIX", "skillos").strip() or "skillos"
    redis_url = os.getenv("SKILLOS_REDIS_URL") or os.getenv("REDIS_URL")
    return CacheConfig(
        enabled=enabled,
        ttl_seconds=ttl_seconds,
        prefix=prefix,
        redis_url=redis_url,
    )


def cache_backend_from_env(config: CacheConfig | None = None) -> CacheBackend | None:
    config = config or cache_config_from_env()
    if not config.enabled:
        return None
    if config.redis_url:
        return RedisCache(config.redis_url)
    return MemoryCache()


def _env_bool(name: str, default: bool) -> bool:
    raw = os.getenv(name)
    if raw is None:
        return default
    normalized = str(raw).strip().lower()
    if normalized in {"1", "true", "yes", "y"}:
        return True
    if normalized in {"0", "false", "no", "n"}:
        return False
    return default


def _env_int(name: str, default: int) -> int:
    raw = os.getenv(name)
    if raw is None:
        return default
    try:
        return int(str(raw).strip())
    except ValueError:
        return default



from __future__ import annotations

from dataclasses import dataclass
import os

from skillos.policy_engine import ApprovalRequirement


def approval_token_from_env() -> str | None:
    return os.getenv("SKILLOS_APPROVAL_TOKEN")


@dataclass(frozen=True)
class ApprovalDecision:
    allowed: bool
    policy_id: str
    status: str


class ApprovalGate:
    def __init__(self, required_token: str | None = None) -> None:
        self._required_token = required_token

    def check(
        self,
        requirement: ApprovalRequirement,
        approval_status: str | None,
        approval_token: str | None = None,
    ) -> ApprovalDecision:
        if not requirement.required:
            return ApprovalDecision(
                allowed=True,
                policy_id="approval_not_required",
                status="not_required",
            )
        if approval_status == "approved":
            if self._required_token and approval_token != self._required_token:
                return ApprovalDecision(
                    allowed=False,
                    policy_id="approval_token_invalid",
                    status="denied",
                )
            return ApprovalDecision(
                allowed=True,
                policy_id="approval_granted",
                status="approved",
            )
        if approval_status == "denied":
            return ApprovalDecision(
                allowed=False,
                policy_id="approval_denied",
                status="denied",
            )
        return ApprovalDecision(
            allowed=False,
            policy_id="approval_required",
            status="required",
        )



from __future__ import annotations

import re

from pydantic import BaseModel, Field, field_validator, model_validator


class SkillEvalCase(BaseModel):
    input: str
    expected: str | None = None
    match: str = Field(default="equals")
    range: tuple[float, float] | None = None
    group: int | None = None

    @model_validator(mode="after")
    def validate_case(self) -> "SkillEvalCase":
        match = self.match.strip().lower()
        self.match = match
        if match == "numeric_range":
            if self.range is None:
                raise ValueError("eval case requires range for numeric_range matcher")
        elif match == "regex_numeric":
            if self.expected is None:
                raise ValueError("eval case requires expected for regex_numeric matcher")
            if self.range is None:
                raise ValueError("eval case requires range for regex_numeric matcher")
            if self.group is not None and self.group < 0:
                raise ValueError("eval case group must be >= 0")
        elif match in {"equals", "contains", "regex"}:
            if self.expected is None:
                raise ValueError("eval case requires expected for matcher")
        else:
            raise ValueError(
                "eval case matcher must be equals|contains|regex|numeric_range|regex_numeric"
            )
        return self


class SkillEvalConfig(BaseModel):
    cases: list[SkillEvalCase] = Field(default_factory=list)
    pass_threshold: float = Field(default=1.0, ge=0.0, le=1.0)
    max_cases: int | None = Field(default=None, ge=1)
    timeout_seconds: float | None = Field(default=None, ge=0.0)
    fail_fast: bool = Field(default=False)


_SEMVER_RE = re.compile(
    r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)"
    r"(?:-[0-9A-Za-z.-]+)?(?:\+[0-9A-Za-z.-]+)?$"
)


class SkillMetadata(BaseModel):
    id: str = Field(..., min_length=3)
    name: str
    description: str
    version: str
    entrypoint: str
    tags: list[str] = Field(default_factory=list)
    execution_mode: str = Field(default="local")
    eval: SkillEvalConfig | None = None
    deprecated: bool = False
    deprecation_reason: str | None = None
    replacement_id: str | None = None

    @field_validator("id")
    @classmethod
    def validate_id(cls, value: str) -> str:
        if "/" not in value or value.startswith("/") or value.endswith("/"):
            raise ValueError("id must be in 'domain/name' format")
        return value

    @field_validator("entrypoint")
    @classmethod
    def validate_entrypoint(cls, value: str) -> str:
        if ":" not in value:
            raise ValueError("entrypoint must be in 'module:function' format")
        return value

    @field_validator("version")
    @classmethod
    def validate_version(cls, value: str) -> str:
        if not _SEMVER_RE.match(value):
            raise ValueError("version must follow semver, e.g. 1.0.0")
        return value

    @field_validator("replacement_id")
    @classmethod
    def validate_replacement_id(cls, value: str | None) -> str | None:
        if value is None:
            return None
        cleaned = value.strip()
        if not cleaned:
            raise ValueError("replacement_id must be in 'domain/name' format")
        if "/" not in cleaned or cleaned.startswith("/") or cleaned.endswith("/"):
            raise ValueError("replacement_id must be in 'domain/name' format")
        return cleaned

    @model_validator(mode="after")
    def validate_deprecation(self) -> "SkillMetadata":
        if self.deprecated and not (self.deprecation_reason or self.replacement_id):
            raise ValueError(
                "deprecated skills must include deprecation_reason or replacement_id"
            )
        return self



from __future__ import annotations
from pathlib import Path
import click

from skillos.skills.paths import default_skills_root
from skillos.skills.registry import SkillRegistry
from skillos.skills.validation import validate_skills


@click.command("add-skill")
@click.argument("skill_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option("--eval/--no-eval", "run_eval", default=False, show_default=True)
@click.option("--eval-required", is_flag=True, default=False)
@click.option(
    "--eval-output",
    "eval_output",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
def add_skill(
    skill_id: str,
    root_path: Path,
    run_eval: bool,
    eval_required: bool,
    eval_output: Path | None,
) -> None:
    """Scaffold a new skill."""
    try:
        from skillos.skills.scaffold import scaffold_skill
        from skillos.skills.errors import SkillValidationError

        metadata_file, implementation_file = scaffold_skill(skill_id, root_path)
    except (FileExistsError, SkillValidationError) as exc:
        raise click.ClickException(str(exc)) from exc

    click.echo(f"Created {metadata_file} and {implementation_file}")

    if run_eval:
        try:
            from skillos.skills.eval import run_skill_eval, save_eval_result, SkillEvalError

            result = run_skill_eval(skill_id, root_path)
            saved_path = save_eval_result(result, root_path, output_path=eval_output)
            click.echo(f"eval_success_rate: {result.success_rate}")
            click.echo(f"eval_saved: {saved_path}")
            if not result.ok:
                raise click.ClickException("eval_failed")
        except SkillEvalError as exc:
            if eval_required:
                raise click.ClickException(str(exc)) from exc
            click.echo(f"eval_skipped: {exc}")


@click.command("deprecate-skill")
@click.argument("skill_id")
@click.option("--reason", default=None)
@click.option("--replacement", "replacement_id", default=None)
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
def deprecate_skill_command(
    skill_id: str,
    reason: str | None,
    replacement_id: str | None,
    root_path: Path,
) -> None:
    """Mark a skill as deprecated."""
    try:
        from skillos.skills.deprecation import deprecate_skill
        from skillos.skills.errors import SkillValidationError

        deprecate_skill(
            root_path,
            skill_id,
            reason=reason,
            replacement_id=replacement_id,
        )
    except (SkillValidationError, FileNotFoundError, ValueError) as exc:
        raise click.ClickException(str(exc)) from exc
    click.echo(f"skill_deprecated: {skill_id}")


@click.command("undeprecate-skill")
@click.argument("skill_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
def undeprecate_skill_command(skill_id: str, root_path: Path) -> None:
    """Remove deprecation flags from a skill."""
    try:
        from skillos.skills.deprecation import undeprecate_skill
        from skillos.skills.errors import SkillValidationError

        undeprecate_skill(root_path, skill_id)
    except (SkillValidationError, FileNotFoundError) as exc:
        raise click.ClickException(str(exc)) from exc
    click.echo(f"skill_undeprecated: {skill_id}")


@click.command("validate")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option(
    "--check-entrypoints/--no-check-entrypoints",
    default=True,
    show_default=True,
)
def validate_command(root_path: Path, check_entrypoints: bool) -> None:
    """Validate all skills under the metadata directory."""
    # validate_skills is imported top-level but it's lightweight logic mostly
    # if it's heavy we can move import inside
    
    issues = validate_skills(root_path, check_entrypoints=check_entrypoints)
    if not issues:
        click.echo("validation_ok")
        return

    for issue in issues:
        skill_hint = issue.skill_id or "unknown"
        click.echo(
            f"invalid_skill: {skill_hint} {issue.category} "
            f"{issue.path} {issue.message}"
        )
    raise click.ClickException("validation_failed")

@click.command("run-skill")
@click.argument("skill_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option("--payload", default="ok", show_default=True)
def run_skill(skill_id: str, root_path: Path, payload: str) -> None:
    """Execute a skill implementation."""
    registry = SkillRegistry(root_path)
    registry.load_all()
    try:
        result = registry.execute(skill_id, payload=payload)
    except Exception as exc:  # pragma: no cover - click handles display
        raise click.ClickException(str(exc)) from exc

    click.echo(result)

@click.command("test")
@click.argument("skill_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option("--payload", default="ok", show_default=True)
@click.option(
    "--coverage-path",
    "coverage_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
def test_skill(
    skill_id: str,
    root_path: Path,
    payload: str,
    coverage_path: Path | None,
) -> None:
    """Run a local test for a skill and generate coverage."""
    try:
        from skillos.testing import run_skill_test

        result = run_skill_test(
            skill_id,
            root_path,
            payload=payload,
            coverage_path=coverage_path,
        )
    except (KeyError, ValueError) as exc:
        raise click.ClickException(str(exc)) from exc

    click.echo(result.output)
    click.echo(f"coverage_written: {result.coverage_path}")


@click.command("eval-skill")
@click.argument("skill_id")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option("--fail-on-threshold/--no-fail-on-threshold", default=True, show_default=True)
@click.option(
    "--output",
    "output_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
@click.option("--save/--no-save", default=True, show_default=True)
def eval_skill(
    skill_id: str,
    root_path: Path,
    fail_on_threshold: bool,
    output_path: Path | None,
    save: bool,
) -> None:
    """Run eval cases defined in skill metadata."""
    try:
        from skillos.skills.eval import run_skill_eval, save_eval_result, SkillEvalError

        result = run_skill_eval(skill_id, root_path)
    except (KeyError, SkillEvalError) as exc:
        raise click.ClickException(str(exc)) from exc

    saved_path = None
    if save:
        saved_path = save_eval_result(result, root_path, output_path=output_path)

    click.echo(f"success_rate: {result.success_rate}")
    click.echo(f"pass_threshold: {result.pass_threshold}")
    click.echo(f"status: {'ok' if result.ok else 'failed'}")
    if saved_path:
        click.echo(f"eval_saved: {saved_path}")
    failed = [case for case in result.cases if not case.passed]
    if failed:
        for case in failed:
            click.echo(
                f"failed_case: input={case.input} match={case.match} details={case.details}"
            )
    if fail_on_threshold and not result.ok:
        raise click.ClickException("eval_failed")



from __future__ import annotations
from pathlib import Path
import click

from skillos.skills.paths import default_skills_root
from skillos.telemetry import EventLogger, default_log_path, default_metrics_path, new_request_id
from skillos.feedback import (
    FeedbackTracker,
    feedback_store_from_env,
    normalize_skill_id,
)
from skillos.experimentation import ExperimentTracker, experiment_store_from_env
from skillos.optimizer import (
    Optimizer, 
    OptimizerConfig, 
    OptimizationLogger, 
    default_optimization_log_path
)
from skillos.skills.registry import SkillRegistry
from skillos.routing import build_router_from_env
from skillos.metrics import load_golden_queries, build_metrics_summary


def _parse_optimization_result(result: str) -> tuple[str, bool]:
    if ":" not in result:
        raise click.ClickException("Result must be in variant:outcome format")
    variant, outcome = result.split(":", 1)
    variant = variant.strip()
    outcome = outcome.strip().lower()
    if not variant:
        raise click.ClickException("Variant id is required")
    if outcome in {"success", "true", "1", "pass", "ok"}:
        return variant, True
    if outcome in {"failure", "false", "0", "fail"}:
        return variant, False
    raise click.ClickException(
        "Outcome must be success/failure or true/false"
    )

@click.command("feedback")
@click.argument("skill_id")
@click.option("--expected-skill-id", default=None)
@click.option("--source", default="cli", show_default=True)
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option(
    "--log-path",
    "log_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
def feedback(
    skill_id: str,
    expected_skill_id: str | None,
    source: str,
    root_path: Path,
    log_path: Path | None,
) -> None:
    """Record feedback to improve routing decisions."""
    tracker = FeedbackTracker(feedback_store_from_env(root_path))
    normalized_skill_id = normalize_skill_id(skill_id)
    normalized_expected = (
        normalize_skill_id(expected_skill_id)
        if expected_skill_id
        else normalized_skill_id
    )
    updates = tracker.record_correction(normalized_skill_id, normalized_expected)
    logger = EventLogger(log_path or default_log_path(root_path), request_id=new_request_id())
    logger.log(
        "feedback_received",
        expected_skill_id=normalized_expected,
        correction=normalized_expected != normalized_skill_id,
        source=source,
        skill_id=normalized_skill_id,
    )
    updated_skills = ", ".join(sorted(updates.keys()))
    click.echo(f"feedback_recorded: {updated_skills}")


@click.command("optimize")
@click.argument("skill_id")
@click.option("--variant", "variants", multiple=True, required=True)
@click.option("--baseline", "baseline_variant", default=None)
@click.option("--result", "results", multiple=True)
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option(
    "--log-path",
    "log_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
@click.option("--min-samples", type=int, default=5, show_default=True)
@click.option("--win-delta", type=float, default=0.1, show_default=True)
@click.option("--rollback-delta", type=float, default=0.1, show_default=True)
def optimize_skill(
    skill_id: str,
    variants: tuple[str, ...],
    baseline_variant: str | None,
    results: tuple[str, ...],
    root_path: Path,
    log_path: Path | None,
    min_samples: int,
    win_delta: float,
    rollback_delta: float,
) -> None:
    """Evaluate A/B experiment results and update active variant."""
    variant_list = [variant.strip() for variant in variants if variant.strip()]
    if len(variant_list) < 2:
        raise click.ClickException("At least two variants are required")
    store = experiment_store_from_env(root_path)
    tracker = ExperimentTracker(
        store,
        skill_id,
        variant_list,
        baseline_variant=baseline_variant,
    )

    for result in results:
        variant_id, success = _parse_optimization_result(result)
        if variant_id not in variant_list:
            raise click.ClickException(f"Unknown variant: {variant_id}")
        tracker.record_outcome(variant_id, success)

    config = OptimizerConfig(
        min_samples=min_samples,
        win_rate_delta=win_delta,
        rollback_delta=rollback_delta,
    )
    logger = OptimizationLogger(log_path or default_optimization_log_path(root_path))
    optimizer = Optimizer(store, logger, config)
    decision = optimizer.evaluate(skill_id)
    click.echo(f"optimization_decision: {decision.action}")


@click.command("metrics")
@click.option(
    "--root",
    "root_path",
    type=click.Path(file_okay=False, path_type=Path),
    default=default_skills_root(),
    show_default=True,
)
@click.option(
    "--golden",
    "golden_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=Path("tests/fixtures/golden_queries.json"),
    show_default=True,
)
@click.option(
    "--output",
    "output_path",
    type=click.Path(dir_okay=False, path_type=Path),
    default=None,
)
def metrics_report(root_path: Path, golden_path: Path, output_path: Path | None) -> None:
    """Generate routing metrics from the golden query set."""
    registry = SkillRegistry(root_path)
    records = registry.load_all()
    feedback_tracker = FeedbackTracker(feedback_store_from_env(root_path))
    router = build_router_from_env(
        [record.metadata for record in records.values()],
        confidence_provider=feedback_tracker.get_confidence,
    )
    golden_queries = load_golden_queries(golden_path)
    summary = build_metrics_summary(router, golden_queries)

    destination = output_path or default_metrics_path(root_path)
    destination.parent.mkdir(parents=True, exist_ok=True)
    destination.write_text(summary, encoding="utf-8")
    click.echo(f"metrics_written: {destination}")



from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from pathlib import Path
import time
from typing import Iterable

from skillos.approval_gate import ApprovalGate, approval_token_from_env
from skillos.authorization import PermissionChecker, default_permissions_path
from skillos.budget import (
    BudgetManager,
    budget_config_from_env,
    budget_usage_store_from_env,
)
from skillos.composition import parallel_limit_from_env
from skillos.circuit_breaker import (
    CircuitBreaker,
    CircuitBreakerStore,
    circuit_breaker_config_from_env,
    default_circuit_breaker_path,
)
from skillos.debugging import StepController
from skillos.skills.deprecation import build_deprecation_warning
from skillos.policy_engine import PolicyEngine, default_policy_path
from skillos.risk_scorer import RiskScorer
from skillos.routing import to_internal_id, to_public_id
from skillos.skills.registry import SkillRegistry
from skillos.telemetry import EventLogger, default_log_path, new_request_id
from skillos.tool_wrapper import ToolWrapper


class PipelineError(ValueError):
    pass


StepGroup = list[str]
StepInput = str | list[str] | tuple[str, ...]


@dataclass(frozen=True)
class PipelineStepResult:
    step_id: str
    status: str
    output: str | None
    duration_ms: float
    order: int
    group: int
    error_class: str | None = None


@dataclass(frozen=True)
class PipelineResult:
    status: str
    output: str | None
    steps: list[PipelineStepResult]
    reason: str | None = None
    warnings: list[dict[str, object]] = field(default_factory=list)


def normalize_pipeline_steps(steps: Iterable[StepInput]) -> list[StepGroup]:
    normalized: list[StepGroup] = []
    for step in steps:
        if isinstance(step, str):
            cleaned = step.strip()
            if not cleaned:
                raise PipelineError("pipeline_steps_required")
            if "|" in cleaned:
                parts = [part.strip() for part in cleaned.split("|") if part.strip()]
                if not parts:
                    raise PipelineError("pipeline_steps_required")
                normalized.append([to_internal_id(part) for part in parts])
            else:
                normalized.append([to_internal_id(cleaned)])
            continue
        if not isinstance(step, (list, tuple)):
            raise PipelineError("pipeline_steps_required")
        if not step:
            raise PipelineError("pipeline_steps_required")
        group: list[str] = []
        for entry in step:
            if not isinstance(entry, str):
                raise PipelineError("pipeline_steps_required")
            cleaned = entry.strip()
            if not cleaned:
                raise PipelineError("pipeline_steps_required")
            group.append(to_internal_id(cleaned))
        normalized.append(group)
    if not normalized:
        raise PipelineError("pipeline_steps_required")
    return normalized


class PipelineRunner:
    def __init__(
        self,
        root_path: Path,
        *,
        log_path: Path | None = None,
        registry: SkillRegistry | None = None,
        budget_manager: BudgetManager | None = None,
        permission_checker: PermissionChecker | None = None,
        tool_wrapper: ToolWrapper | None = None,
        parallel_limit: int | None = None,
        step_controller: StepController | None = None,
    ) -> None:
        self.root_path = Path(root_path)
        self.log_path = log_path or default_log_path(self.root_path)
        self.registry = registry or SkillRegistry(self.root_path)
        if registry is None:
            self.registry.load_all()

        self.budget_manager = budget_manager or BudgetManager(
            budget_usage_store_from_env(self.root_path),
            budget_config_from_env(),
        )
        self.permission_checker = permission_checker or PermissionChecker.from_path(
            default_permissions_path(self.root_path)
        )
        policy_engine = PolicyEngine.from_path(default_policy_path(self.root_path))
        self.tool_wrapper = tool_wrapper or ToolWrapper(
            risk_scorer=RiskScorer(),
            policy_engine=policy_engine,
            approval_gate=ApprovalGate(required_token=approval_token_from_env()),
        )
        self.circuit_breaker = CircuitBreaker(
            CircuitBreakerStore(default_circuit_breaker_path(self.root_path)),
            circuit_breaker_config_from_env(),
        )
        self.parallel_limit = parallel_limit or parallel_limit_from_env()
        self.step_controller = step_controller

    def run(
        self,
        steps: Iterable[StepInput],
        *,
        payload: str = "ok",
        approval_status: str | None = None,
        approval_token: str | None = None,
        role: str | None = None,
        attributes: dict[str, object] | None = None,
        request_id: str | None = None,
        session_context: dict[str, object] | None = None,
    ) -> PipelineResult:
        self._maybe_reload_registry()
        normalized_steps = normalize_pipeline_steps(steps)
        request_id = request_id or new_request_id()
        logger = EventLogger(self.log_path, request_id=request_id)

        current_payload = payload
        step_results: list[PipelineStepResult] = []
        warnings: list[dict[str, object]] = []
        order_index = 0
        for group_index, group in enumerate(normalized_steps):
            if len(group) == 1:
                step_id = group[0]
                public_id = to_public_id(step_id)
                block = self._authorize_step(
                    logger,
                    step_id,
                    public_id,
                    current_payload,
                    role,
                    attributes,
                    approval_status,
                    approval_token,
                    order_index,
                    group_index,
                    warnings,
                )
                if block:
                    step_results.append(block)
                    return PipelineResult(
                        status=block.status,
                        output=None,
                        steps=step_results,
                        reason=block.error_class,
                        warnings=warnings,
                    )
                step_result = self._execute_step(
                    logger,
                    step_id,
                    public_id,
                    current_payload,
                    order_index,
                    group_index,
                    role=role,
                    attributes=attributes,
                    approval_status=approval_status,
                    approval_token=approval_token,
                    session_context=session_context,
                )
                step_results.append(step_result)
                if step_result.status != "success":
                    return PipelineResult(
                        status=step_result.status,
                        output=None,
                        steps=step_results,
                        reason=step_result.error_class,
                        warnings=warnings,
                    )
                current_payload = step_result.output or ""
            else:
                group_result = self._execute_parallel_group(
                    logger,
                    group,
                    current_payload,
                    role,
                    attributes,
                    approval_status,
                    approval_token,
                    order_index,
                    group_index,
                    warnings,
                    session_context=session_context,
                )
                step_results.extend(group_result.steps)
                if group_result.status != "success":
                    return PipelineResult(
                        status=group_result.status,
                        output=None,
                        steps=step_results,
                        reason=group_result.reason,
                        warnings=warnings,
                    )
                current_payload = group_result.output or ""
            order_index += len(group)

        return PipelineResult(
            status="success",
            output=current_payload,
            steps=step_results,
            reason=None,
            warnings=warnings,
        )

    async def run_async(
        self,
        steps: Iterable[StepInput],
        *,
        payload: str = "ok",
        approval_status: str | None = None,
        approval_token: str | None = None,
        role: str | None = None,
        attributes: dict[str, object] | None = None,
        request_id: str | None = None,
        session_context: dict[str, object] | None = None,
    ) -> PipelineResult:
        self._maybe_reload_registry()
        normalized_steps = normalize_pipeline_steps(steps)
        request_id = request_id or new_request_id()
        logger = EventLogger(self.log_path, request_id=request_id)

        current_payload = payload
        step_results: list[PipelineStepResult] = []
        warnings: list[dict[str, object]] = []
        order_index = 0
        
        try:
            for group_index, group in enumerate(normalized_steps):
                if len(group) == 1:
                    step_id = group[0]
                    public_id = to_public_id(step_id)
                    block = self._authorize_step(
                        logger,
                        step_id,
                        public_id,
                        current_payload,
                        role,
                        attributes,
                        approval_status,
                        approval_token,
                        order_index,
                        group_index,
                        warnings,
                    )
                    if block:
                        step_results.append(block)
                        return PipelineResult(
                            status=block.status,
                            output=None,
                            steps=step_results,
                            reason=block.error_class,
                            warnings=warnings,
                        )
                    
                    step_result = await self._execute_step_async(
                        logger,
                        step_id,
                        public_id,
                        current_payload,
                        order_index,
                        group_index,
                        role=role,
                        attributes=attributes,
                        approval_status=approval_status,
                        approval_token=approval_token,
                        session_context=session_context,
                    )
                    step_results.append(step_result)
                    if step_result.status != "success":
                        return PipelineResult(
                            status=step_result.status,
                            output=None,
                            steps=step_results,
                            reason=step_result.error_class,
                            warnings=warnings,
                        )
                    current_payload = step_result.output or ""
                else:
                    group_result = await self._execute_parallel_group_async(
                        logger,
                        group,
                        current_payload,
                        role,
                        attributes,
                        approval_status,
                        approval_token,
                        order_index,
                        group_index,
                        warnings,
                        session_context=session_context,
                    )
                    step_results.extend(group_result.steps)
                    if group_result.status != "success":
                        return PipelineResult(
                            status=group_result.status,
                            output=None,
                            steps=step_results,
                            reason=group_result.reason,
                            warnings=warnings,
                        )
                    current_payload = group_result.output or ""
                order_index += len(group)
        except asyncio.CancelledError:
            logger.log("pipeline_cancelled", request_id=request_id)
            # Resources are cleaned up by individual step cancellation handlers if needed
            raise

        return PipelineResult(
            status="success",
            output=current_payload,
            steps=step_results,
            reason=None,
            warnings=warnings,
        )

    def _authorize_step(
        self,
        logger: EventLogger,
        internal_id: str,
        public_id: str,
        payload: str,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        order_index: int,
        group_index: int,
        warnings: list[dict[str, object]],
    ) -> PipelineStepResult | None:
        start = time.perf_counter()
        metadata = self.registry.get(internal_id)
        self._warn_deprecated(metadata, public_id, warnings, logger=logger)

        permission_decision = self._permission_check(
            internal_id,
            role,
            attributes,
            metadata,
            logger=logger,
        )
        if not permission_decision.allowed:
            return self._blocked_step(
                logger,
                public_id,
                order_index,
                group_index,
                start,
                payload,
                permission_decision.policy_id,
            )

        decision = self._approval_check(
            internal_id,
            payload,
            approval_status,
            approval_token,
            logger=logger,
        )
        if not decision.approval.allowed:
            return self._blocked_step(
                logger,
                public_id,
                order_index,
                group_index,
                start,
                payload,
                decision.approval.policy_id,
            )

        allowed, reason = self._circuit_check(internal_id, logger=logger)
        if not allowed:
            return self._blocked_step(
                logger,
                public_id,
                order_index,
                group_index,
                start,
                payload,
                reason or "circuit_breaker_open",
            )

        budget_result = self._budget_check(logger=logger)
        if not budget_result.allowed:
            logger.log(
                "policy_decision",
                allowed=False,
                policy_id="budget_block",
            )
            return self._blocked_step(
                logger,
                public_id,
                order_index,
                group_index,
                start,
                payload,
                budget_result.reason or "budget_block",
            )
        return None

    def _blocked_step(
        self,
        logger: EventLogger,
        step_id: str,
        order_index: int,
        group_index: int,
        start: float,
        payload: str,
        reason: str,
    ) -> PipelineStepResult:
        duration_ms = (time.perf_counter() - start) * 1000
        logger.log(
            "execution_result",
            status="blocked",
            duration_ms=duration_ms,
            error_class=reason,
            skill_id=step_id,
        )
        result = PipelineStepResult(
            step_id=step_id,
            status="blocked",
            output=None,
            duration_ms=duration_ms,
            order=order_index,
            group=group_index,
            error_class=reason,
        )
        self._log_pipeline_step(logger, result)
        self._pause_step(
            step_id,
            inputs={"payload": payload},
            outputs={"status": "blocked", "reason": reason},
            duration_ms=duration_ms,
        )
        return result

    def _execute_step(
        self,
        logger: EventLogger,
        internal_id: str,
        public_id: str,
        payload: str,
        order_index: int,
        group_index: int,
        *,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        pause: bool = True,
        session_context: dict[str, object] | None = None,
    ) -> PipelineStepResult:
        start = time.perf_counter()
        try:
            output = self.registry.execute(
                internal_id,
                payload=payload,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                session_context=session_context,
                charge_budget=False,
            )
        except Exception as exc:
            import traceback
            print(f"DEBUG: EXECUTION ERROR for {internal_id}: {exc}")
            traceback.print_exc()
            duration_ms = (time.perf_counter() - start) * 1000
            logger.log(
                "execution_result",
                status="error",
                duration_ms=duration_ms,
                error_class=exc.__class__.__name__,
                skill_id=public_id,
            )
            self.circuit_breaker.record_failure(internal_id)
            result = PipelineStepResult(
                step_id=public_id,
                status="error",
                output=None,
                duration_ms=duration_ms,
                order=order_index,
                group=group_index,
                error_class=exc.__class__.__name__,
            )
            self._log_pipeline_step(logger, result)
            if pause:
                self._pause_step(
                    public_id,
                    inputs={"payload": payload},
                    outputs={
                        "status": "error",
                        "error_class": exc.__class__.__name__,
                    },
                    duration_ms=duration_ms,
                )
            return result

        duration_ms = (time.perf_counter() - start) * 1000
        logger.log(
            "execution_result",
            status="success",
            duration_ms=duration_ms,
            skill_id=public_id,
        )
        self.circuit_breaker.record_success(internal_id)
        result = PipelineStepResult(
            step_id=public_id,
            status="success",
            output=str(output),
            duration_ms=duration_ms,
            order=order_index,
            group=group_index,
        )
        self._log_pipeline_step(logger, result)
        if pause:
            self._pause_step(
                public_id,
                inputs={"payload": payload},
                outputs={"status": "success", "output": str(output)},
                duration_ms=duration_ms,
            )
        return result

    async def _execute_step_async(
        self,
        logger: EventLogger,
        internal_id: str,
        public_id: str,
        payload: str,
        order_index: int,
        group_index: int,
        *,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        pause: bool = True,
        session_context: dict[str, object] | None = None,
    ) -> PipelineStepResult:
        start = time.perf_counter()
        try:
            output = await self.registry.execute_async(
                internal_id,
                payload=payload,
                role=role,
                attributes=attributes,
                approval_status=approval_status,
                approval_token=approval_token,
                session_context=session_context,
                charge_budget=False,
            )
        except asyncio.CancelledError:
            # Re-raise to let the runner handle it
            raise
        except Exception as exc:
            duration_ms = (time.perf_counter() - start) * 1000
            logger.log(
                "execution_result",
                status="error",
                duration_ms=duration_ms,
                error_class=exc.__class__.__name__,
                skill_id=public_id,
            )
            self.circuit_breaker.record_failure(internal_id)
            result = PipelineStepResult(
                step_id=public_id,
                status="error",
                output=None,
                duration_ms=duration_ms,
                order=order_index,
                group=group_index,
                error_class=exc.__class__.__name__,
            )
            self._log_pipeline_step(logger, result)
            if pause:
                self._pause_step(
                    public_id,
                    inputs={"payload": payload},
                    outputs={
                        "status": "error",
                        "error_class": exc.__class__.__name__,
                    },
                    duration_ms=duration_ms,
                )
            return result

        duration_ms = (time.perf_counter() - start) * 1000
        logger.log(
            "execution_result",
            status="success",
            duration_ms=duration_ms,
            skill_id=public_id,
        )
        self.circuit_breaker.record_success(internal_id)
        result = PipelineStepResult(
            step_id=public_id,
            status="success",
            output=str(output),
            duration_ms=duration_ms,
            order=order_index,
            group=group_index,
        )
        self._log_pipeline_step(logger, result)
        if pause:
            self._pause_step(
                public_id,
                inputs={"payload": payload},
                outputs={"status": "success", "output": str(output)},
                duration_ms=duration_ms,
            )
        return result

    def _execute_parallel_group(
        self,
        logger: EventLogger,
        group: list[str],
        payload: str,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        order_index: int,
        group_index: int,
        warnings: list[dict[str, object]],
        session_context: dict[str, object] | None = None,
    ) -> PipelineResult:
        step_results: list[PipelineStepResult] = []
        public_ids = [to_public_id(step_id) for step_id in group]

        for offset, (internal_id, public_id) in enumerate(zip(group, public_ids)):
            block = self._authorize_step(
                logger,
                internal_id,
                public_id,
                payload,
                role,
                attributes,
                approval_status,
                approval_token,
                order_index + offset,
                group_index,
                warnings,
            )
            if block:
                step_results.append(block)
                return PipelineResult(
                    status=block.status,
                    output=None,
                    steps=step_results,
                    reason=block.error_class,
                )

        outputs: list[str | None] = [None for _ in group]
        results: list[PipelineStepResult | None] = [None for _ in group]
        with ThreadPoolExecutor(max_workers=self.parallel_limit) as executor:
            future_map = {}
            for offset, (internal_id, public_id) in enumerate(zip(group, public_ids)):
                future = executor.submit(
                    self._execute_step,
                    logger,
                    internal_id,
                    public_id,
                    payload,
                    order_index + offset,
                    group_index,
                    role=role,
                    attributes=attributes,
                    approval_status=approval_status,
                    approval_token=approval_token,
                    pause=False,
                    session_context=session_context,
                )
                future_map[future] = offset
            for future in as_completed(future_map):
                index = future_map[future]
                result = future.result()
                results[index] = result
                outputs[index] = result.output
                if result.status != "success":
                    for pending in future_map:
                        if pending is not future:
                            pending.cancel()
                    step_results.extend(result for result in results if result)
                    self._pause_parallel_group(payload, results)
                    return PipelineResult(
                        status=result.status,
                        output=None,
                        steps=step_results,
                        reason=result.error_class,
                    )

        step_results.extend(result for result in results if result)
        self._pause_parallel_group(payload, results)
        combined = "\n".join(
            output for output in outputs if output is not None
        )
        return PipelineResult(
            status="success",
            output=combined,
            steps=step_results,
            reason=None,
        )

    async def _execute_parallel_group_async(
        self,
        logger: EventLogger,
        group: list[str],
        payload: str,
        role: str | None,
        attributes: dict[str, object] | None,
        approval_status: str | None,
        approval_token: str | None,
        order_index: int,
        group_index: int,
        warnings: list[dict[str, object]],
        session_context: dict[str, object] | None = None,
    ) -> PipelineResult:
        step_results: list[PipelineStepResult] = []
        public_ids = [to_public_id(step_id) for step_id in group]

        # Auth check remains sync as it's quick metadata/policy check
        for offset, (internal_id, public_id) in enumerate(zip(group, public_ids)):
            block = self._authorize_step(
                logger,
                internal_id,
                public_id,
                payload,
                role,
                attributes,
                approval_status,
                approval_token,
                order_index + offset,
                group_index,
                warnings,
            )
            if block:
                step_results.append(block)
                return PipelineResult(
                    status=block.status,
                    output=None,
                    steps=step_results,
                    reason=block.error_class,
                )

        task_map: dict[asyncio.Task[PipelineStepResult], int] = {}
        for offset, (internal_id, public_id) in enumerate(zip(group, public_ids)):
            task = asyncio.create_task(
                self._execute_step_async(
                    logger,
                    internal_id,
                    public_id,
                    payload,
                    order_index + offset,
                    group_index,
                    role=role,
                    attributes=attributes,
                    approval_status=approval_status,
                    approval_token=approval_token,
                    pause=False,
                    session_context=session_context,
                )
            )
            task_map[task] = offset

        results: list[PipelineStepResult | None] = [None for _ in group]
        outputs: list[str | None] = [None for _ in group]

        pending: set[asyncio.Task[PipelineStepResult]] = set(task_map.keys())
        while pending:
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )
            for task in done:
                index = task_map[task]
                try:
                    res = task.result()
                except asyncio.CancelledError:
                    continue
                except Exception as exc:
                    res = PipelineStepResult(
                        step_id=public_ids[index],
                        status="error",
                        output=None,
                        duration_ms=0.0,
                        order=order_index + index,
                        group=group_index,
                        error_class=exc.__class__.__name__,
                    )
                    self._log_pipeline_step(logger, res)
                results[index] = res
                outputs[index] = res.output
                if res.status != "success":
                    for pending_task in pending:
                        pending_task.cancel()
                    if pending:
                        await asyncio.gather(*pending, return_exceptions=True)
                    step_results.extend([r for r in results if r])
                    self._pause_parallel_group(payload, results)
                    return PipelineResult(
                        status=res.status,
                        output=None,
                        steps=step_results,
                        reason=res.error_class,
                    )

        step_results.extend([r for r in results if r])
        self._pause_parallel_group(payload, results)
        combined = "\n".join(output for output in outputs if output is not None)
        return PipelineResult(
            status="success",
            output=combined,
            steps=step_results,
            reason=None,
        )

    def _circuit_check(
        self,
        internal_id: str,
        *,
        logger: EventLogger,
    ) -> tuple[bool, str | None]:
        circuit_decision = self.circuit_breaker.allow(internal_id)
        if circuit_decision.allowed:
            return True, None
        logger.log(
            "policy_decision",
            allowed=False,
            policy_id="circuit_breaker_open",
        )
        return False, circuit_decision.reason or "circuit_breaker_open"

    def _budget_check(self, *, logger: EventLogger):
        budget_result = self.budget_manager.authorize()
        logger.log(
            "budget_check",
            allowed=budget_result.allowed,
            reason=budget_result.reason,
            model=budget_result.model,
            estimated_cost=budget_result.estimated_cost,
            remaining_daily=budget_result.remaining_daily,
            remaining_monthly=budget_result.remaining_monthly,
        )
        return budget_result

    @staticmethod
    def _warn_deprecated(
        metadata,
        public_id: str,
        warnings: list[dict[str, object]],
        *,
        logger: EventLogger,
    ) -> None:
        if not metadata or not metadata.deprecated:
            return
        warnings.append(build_deprecation_warning(public_id, metadata))
        logger.log(
            "deprecated_skill_used",
            skill_id=public_id,
            deprecation_reason=metadata.deprecation_reason,
            replacement_id=metadata.replacement_id,
        )

    def _permission_check(
        self,
        internal_id: str,
        role: str | None,
        attributes: dict[str, object] | None,
        metadata,
        *,
        logger: EventLogger,
    ):
        skill_tags = metadata.tags if metadata else None
        permission_decision = self.permission_checker.authorize(
            internal_id,
            role,
            skill_tags=skill_tags,
            attributes=attributes,
        )
        logger.log(
            "permission_decision",
            allowed=permission_decision.allowed,
            policy_id=permission_decision.policy_id,
            role=permission_decision.role,
            skill_id=permission_decision.skill_id,
            required_permissions=permission_decision.required_permissions,
            missing_permissions=permission_decision.missing_permissions,
        )
        return permission_decision

    def _approval_check(
        self,
        internal_id: str,
        payload: str,
        approval_status: str | None,
        approval_token: str | None,
        *,
        logger: EventLogger,
    ):
        decision = self.tool_wrapper.authorize(
            internal_id,
            payload,
            approval_status=approval_status,
            approval_token=approval_token,
        )
        logger.log(
            "policy_decision",
            allowed=decision.approval.allowed,
            policy_id=decision.approval.policy_id,
            approval_required=decision.requirement.required,
            approval_reason=decision.requirement.policy_id,
            risk_score=decision.risk.score,
            risk_level=decision.risk.level,
            approval_status=decision.approval.status,
        )
        return decision

    @staticmethod
    def _log_pipeline_step(logger: EventLogger, result: PipelineStepResult) -> None:
        payload = {
            "status": result.status,
            "step_id": result.step_id,
            "duration_ms": result.duration_ms,
            "order": result.order,
            "group": result.group,
        }
        if result.error_class:
            payload["error_class"] = result.error_class
        logger.log("pipeline_step", **payload)

    def _pause_step(
        self,
        step_id: str,
        *,
        inputs: dict[str, object] | None,
        outputs: dict[str, object] | None,
        duration_ms: float | None,
    ) -> None:
        if not self.step_controller:
            return
        self.step_controller.pause(
            step_id,
            inputs=inputs,
            outputs=outputs,
            duration_ms=duration_ms,
        )

    def _pause_parallel_group(
        self,
        payload: str,
        results: list[PipelineStepResult | None],
    ) -> None:
        if not self.step_controller:
            return
        for result in results:
            if not result:
                continue
            outputs = {"status": result.status}
            if result.error_class:
                outputs["error_class"] = result.error_class
            if result.output is not None:
                outputs["output"] = result.output
            self._pause_step(
                result.step_id,
                inputs={"payload": payload},
                outputs=outputs,
                duration_ms=result.duration_ms,
            )

    def _maybe_reload_registry(self) -> None:
        self.registry.reload_if_changed()



from __future__ import annotations
from pathlib import Path
import click

try:
    from skillos.jwt_auth import jwt_token_from_env, decode_jwt, jwt_config_from_env, JwtValidationError, JwtClaims
except ImportError:
    # Handle cases where dependencies might not be logically loaded yet, though unlikely in running app
    pass

def _emit_warnings(warnings: list[dict[str, object]] | None) -> None:
    if not warnings:
        return
    for warning in warnings:
        message = warning.get("message") or warning.get("code") or "warning"
        click.echo(f"warning: {message}", err=True)

def _blocked_message(reason: str) -> str:
    budget_reasons = {
        "per_request_limit_exceeded",
        "daily_limit_exceeded",
        "monthly_limit_exceeded",
    }
    if reason in budget_reasons:
        return f"budget_exceeded: {reason}"
    return reason

def _approval_message(policy_id: str) -> str:
    if policy_id == "approval_denied":
        return "approval_denied"
    if policy_id == "approval_token_invalid":
        return "approval_token_invalid"
    return "approval_required"

def _resolve_auth_context(jwt_token: str | None) -> JwtClaims | None:
    # Importing here to potentially avoid circular imports if jwt_auth depends on something unusual
    # But standard imports are generally preferred now if no cycles. 
    # For safety against cycles/heavy load if jwt_auth is heavy:
    from skillos.jwt_auth import jwt_token_from_env, decode_jwt, jwt_config_from_env, JwtValidationError

    token = jwt_token or jwt_token_from_env()
    if not token:
        return None
    try:
        return decode_jwt(token, jwt_config_from_env())
    except JwtValidationError as exc:
        raise click.ClickException(f"jwt_invalid: {exc}") from exc

def _apply_auth_context(
    root_path: Path,
    role: str | None,
    jwt_token: str | None,
) -> tuple[Path, str | None, dict[str, object] | None]:
    from skillos.tenancy import tenant_id_from_path, resolve_tenant_root

    context = _resolve_auth_context(jwt_token)
    if context is None:
        return root_path, role, None
    root_tenant = tenant_id_from_path(root_path)
    if context.tenant_id and root_tenant and root_tenant != context.tenant_id:
        raise click.ClickException("tenant_mismatch")
    resolved_root = root_path
    if context.tenant_id and not root_tenant:
        resolved_root = resolve_tenant_root(root_path, tenant_id=context.tenant_id)
    resolved_role = role or context.role
    return resolved_root, resolved_role, context.attributes



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Mapping

@dataclass(frozen=True)
class AppConfig:
    env: str
    log_level: str
    postgres_dsn: str | None
    redis_url: str | None
    
    @classmethod
    def from_env(cls, env: Mapping[str, str] | None = None) -> "AppConfig":
        env_map = env or os.environ
        return cls(
            env=env_map.get("SKILLOS_ENV", "dev").lower(),
            log_level=env_map.get("SKILLOS_LOG_LEVEL", "INFO").upper(),
            postgres_dsn=env_map.get("SKILLOS_POSTGRES_DSN"),
            redis_url=env_map.get("SKILLOS_REDIS_URL") or env_map.get("REDIS_URL"),
        )
    
    @property
    def is_prod(self) -> bool:
        return self.env in ("prod", "production")

def get_config() -> AppConfig:
    return AppConfig.from_env()
